"""
Dashboard for the qual page
"""

from powa.dashboards import (
    ContentWidget,
    Dashboard,
    DashboardPage,
    Grid,
    MetricDef,
    MetricGroupDef,
)
from powa.query import QueryOverview
from powa.sql import qual_constants, resolve_quals
from powa.sql.views import qualstat_getstatdata
from tornado.web import HTTPError


class QualConstantsMetricGroup(MetricGroupDef):
    """
    Metric group used for the qual charts.
    """

    name = "QualConstants"
    data_url = r"/server/(\d+)/metrics/database/([^\/]+)/query/(-?\d+)/qual/(\d+)/constants"
    xaxis = "rownumber"
    occurences = MetricDef(label="Occurences")
    grouper = "constants"

    @property
    def query(self):
        most_used = qual_constants(
            "%(server)s",
            "most_used",
            """
            datname = %(database)s AND
            coalesce_range && tstzrange(%(from)s, %(to)s)""",
            "%(query)s",
            "%(qual)s",
            top=10,
        )

        correlated = qualstat_getstatdata(
            extra_where=["qualid = %(qual)s", "queryid = %(query)s"]
        )
        sql = """SELECT sub.*, correlated.occurences as total_occurences
            FROM (
                SELECT *
                FROM (
                    {most_used}
                ) AS most_used
            ) AS sub, (
                {correlated}
            ) AS correlated
        """.format(most_used=most_used, correlated=correlated)

        return sql

    def add_params(self, params):
        params["queryids"] = [int(params["query"])]
        return params

    def post_process(self, data, server, database, query, qual, **kwargs):
        if not data["data"]:
            return data
        max_rownumber = 0
        total_top10 = 0
        total = None
        d = {"total_occurences": 0}
        for d in data["data"]:
            max_rownumber = max(max_rownumber, d["rownumber"])
            total_top10 += d["occurences"]
        else:
            total = d["total_occurences"]
        data["data"].append(
            {
                "occurences": total - total_top10,
                "rownumber": max_rownumber + 1,
                "constants": "Others",
            }
        )
        return data


class QualDetail(ContentWidget):
    """
    Content widget showing detail for a specific qual.
    """

    title = "Detail for this Qual"
    data_url = (
        r"/server/(\d+)/database/([^\/]+)/query/(-?\d+)/qual/(\d+)/detail"
    )

    def get(self, server, database, query, qual):
        try:
            # Check remote access first
            remote_conn = self.connect(
                server, database=database, remote_access=True
            )
        except Exception as e:
            raise HTTPError(
                501, "Could not connect to remote server: %s" % str(e)
            )
        stmt = qualstat_getstatdata(
            extra_select=["queryid = %(query)s AS is_my_query"],
            extra_where=["qualid = %(qualid)s", "occurences > 0"],
            extra_groupby=["queryid"],
        )
        quals = list(
            self.execute(
                stmt,
                params={
                    "server": server,
                    "query": query,
                    "from": self.get_argument("from"),
                    "to": self.get_argument("to"),
                    "queryids": [query],
                    "qualid": qual,
                },
            )
        )

        my_qual = None

        for qual in quals:
            if qual["is_my_query"]:
                my_qual = resolve_quals(remote_conn, [qual])[0]

        if my_qual is None:
            self.render_json(None)
            return

        self.render_json(my_qual)


class OtherQueriesMetricGroup(MetricGroupDef):
    """Metric group showing other queries for this qual."""

    name = "other_queries"
    xaxis = "queryid"
    axis_type = "category"
    data_url = r"/server/(\d+)/metrics/database/([^\/]+)/query/(-?\d+)/qual/(\d+)/other_queries"
    query_str = MetricDef(label="Query", type="query", url_attr="url")

    @property
    def query(self):
        return """
            SELECT distinct queryid, query,
            query as query_str, pd.srvid
            FROM {powa}.powa_qualstats_quals pqs
            JOIN {powa}.powa_statements USING (queryid, dbid, srvid, userid)
            JOIN {powa}.powa_databases pd ON pd.oid = pqs.dbid AND pd.srvid =
            pqs.srvid
            WHERE qualid = %(qual)s
                AND pqs.queryid != %(query)s
                AND pd.srvid = %(server)s
                AND pd.datname = %(database)s"""

    def process(self, val, database=None, **kwargs):
        val["url"] = self.reverse_url(
            "QueryOverview", val["srvid"], database, val["queryid"]
        )
        return val


class QualOverview(DashboardPage):
    """
    Dashboard page for a specific qual.
    """

    base_url = r"/server/(\d+)/database/([^\/]+)/query/(-?\d+)/qual/(\d+)"
    params = ["server", "database", "query", "qual"]
    datasources = [
        QualDetail,
        OtherQueriesMetricGroup,
        QualConstantsMetricGroup,
    ]
    parent = QueryOverview
    title = "Predicate Overview"

    def dashboard(self):
        # This COULD be initialized in the constructor, but tornado < 3 doesn't
        # call it
        if getattr(self, "_dashboard", None) is not None:
            return self._dashboard

        self._dashboard = Dashboard(
            "Qual %(qual)s",
            [
                [QualDetail],
                [
                    Grid(
                        "Other queries",
                        metrics=OtherQueriesMetricGroup.all(),
                        columns=[],
                    )
                ],
                [
                    Grid(
                        "Most executed values",
                        metrics=[QualConstantsMetricGroup.occurences],
                        x_label_attr="constants",
                        renderer="distribution",
                    )
                ],
            ],
        )

        return self._dashboard



"""
Dashboard for the powa-collector summary page, and other infrastructure for the
collector handling.
"""

from __future__ import absolute_import

import json
from powa.dashboards import MetricGroupDef
from powa.framework import AuthHandler


class CollectorServerDetail(MetricGroupDef):
    name = "Server details"
    data_url = r"/config/(\d+)/server_details/"
    params = ["server"]

    @property
    def query(self):
        return None

    def post_process(self, data, server, **kwargs):
        sql = """SELECT *
            FROM {powa}.powa_servers s
            JOIN {powa}.powa_snapshot_metas m ON m.srvid = s.id
            WHERE s.id = %(server)s"""

        row = self.execute(sql, params={"server": server})

        # unexisting server, bail out
        if len(row) != 1:
            data["messages"] = {"alert": ["This server does not exists"]}
            return data

        status = "unknown"
        if server == "0":
            status = self.execute("""SELECT
                    CASE WHEN count(*) = 1 THEN 'running'
                    ELSE 'stopped'
                    END AS status
                    FROM pg_stat_activity
                    WHERE application_name LIKE 'PoWA - %%'""")[0]["status"]
        else:
            raw = self.notify_collector("WORKERS_STATUS", [server], 2)

            status = None
            # did we receive a valid answer?
            if len(raw) != 0 and "OK" in raw[0]:
                # just keep the first one
                tmp = raw[0]["OK"]
                if server in tmp:
                    status = json.loads(tmp)[server]

        if status is None:
            return {
                "messages": {
                    "warning": ["Could not get status for this instance"]
                },
                "data": [],
            }

        msg = "Collector status for this instance: " + status
        level = "alert"
        if status == "running":
            level = "success"

        return {"messages": {level: [msg]}, "data": []}


class CollectorReloadHandler(AuthHandler):
    """Page allowing to choose a server."""

    def get(self):
        res = False

        answers = self.notify_collector("RELOAD")

        # iterate over the results.  If at least one OK is received, report
        # success, otherwise failure
        for a in answers:
            if "OK" in a:
                res = True
                break

        self.render_json(res)


class CollectorForceSnapshotHandler(AuthHandler):
    """Request an immediate snapshot on the given server."""

    def get(self, server):
        answers = self.notify_collector("FORCE_SNAPSHOT", [server])

        self.render_json(answers)


class CollectorDbCatRefreshHandler(AuthHandler):
    """Refresh the catalogs for the given cadatabase(s) on the given server."""

    def post(self):
        payload = json.loads(self.request.body.decode("utf8"))
        nb_db = len(payload["dbnames"])
        args = [payload["srvid"], str(nb_db)]
        if nb_db > 0:
            args.extend(payload["dbnames"])

        answers = self.notify_collector("REFRESH_DB_CAT", args)

        self.render_json(answers)



"""
Utilities for the basis of Powa
"""

import logging
import pickle
import psycopg2
import random
import re
import select
import time
from collections import defaultdict
from powa import ui_methods
from powa.json import JSONizable, to_json
from psycopg2.extensions import connection as _connection
from psycopg2.extensions import cursor as _cursor
from psycopg2.extras import RealDictCursor
from tornado.options import options
from tornado.web import HTTPError, RequestHandler, authenticated


class CustomConnection(_connection):
    """
    Custom psycopg2 connection class that takes care of expanding extension
    schema and optionally logs various information at debug level, both on
    successful execution and in case of error.

    Supports either plain cursor (through CustomCursor) or RealDictCursor
    (through CustomDictCursor).

    Before execution, and if a _nsps object is found cached in the connection,
    the query will be formatted using this _nsps dict, which contains a list of
    extension_name -> escaped_schema_name mapping.
    All you need to do is pass query strings of the form
    SELECT ... FROM {extension_name}.some_relation ...
    """

    def initialize(self, logger, srvid, dsn, encoding_query, debug):
        self._logger = logger
        self._srvid = srvid or 0
        self._dsn = dsn
        self._debug = debug

        if encoding_query is not None:
            self.set_client_encoding(encoding_query["client_encoding"])

    def cursor(self, *args, **kwargs):
        factory = kwargs.get("cursor_factory")

        if factory is None:
            kwargs["cursor_factory"] = CustomCursor
        elif factory == RealDictCursor:
            kwargs["cursor_factory"] = CustomDictCursor
        else:
            msg = "Unsupported cursor_factory: %s" % factory.__name__
            self._logger.error(msg)
            raise Exception(msg)

        return _connection.cursor(self, *args, **kwargs)


class CustomDictCursor(RealDictCursor):
    def execute(self, query, params=None):
        query = resolve_nsps(query, self.connection)

        self.timestamp = time.time()
        try:
            return super(CustomDictCursor, self).execute(query, params)
        except Exception as e:
            log_query(self, query, params, e)
            raise e
        finally:
            log_query(self, query, params)


class CustomCursor(_cursor):
    def execute(self, query, params=None):
        query = resolve_nsps(query, self.connection)

        self.timestamp = time.time()
        try:
            return super(CustomCursor, self).execute(query, params)
        except Exception as e:
            log_query(self, query, params, e)
        finally:
            log_query(self, query, params)


def resolve_nsps(query, connection):
    try:
        if hasattr(connection, "_nsps"):
            return query.format(**connection._nsps)
    except KeyError as e:
        raise Exception("Extension not found: " + e.args[0])

    return query


def log_query(cls, query, params=None, exception=None):
    t = round((time.time() - cls.timestamp) * 1000, 2)

    fmt = ""
    if exception is not None:
        fmt = "Error during query execution:\n{}\n".format(exception)

    fmt += "query on {dsn} (srvid {srvid}): {ms} ms\n{query}"
    if params is not None:
        fmt += "\n{params}"

    cls.connection._logger.debug(
        fmt.format(
            ms=t,
            query=query,
            params=params,
            dsn=cls.connection._dsn,
            srvid=cls.connection._srvid,
        )
    )


class BaseHandler(RequestHandler, JSONizable):
    """
    Subclass of Tornado RequestHandler adding a bunch
    of utility methods.
    """

    def __init__(self, *args, **kwargs):
        super(BaseHandler, self).__init__(*args, **kwargs)
        self.flashed_messages = {}
        self._databases = None
        self._servers = None
        self._connections = {}
        self._ext_versions = defaultdict(lambda: defaultdict(dict))
        self.url_prefix = options.url_prefix
        self.logger = logging.getLogger("tornado.application")
        if self.application.settings["debug"]:
            self.logger.setLevel(logging.DEBUG)

    def __get_url(self, **connoptions):
        url = " ".join(
            [
                "%s=%s" % (k, v)
                for (k, v) in connoptions.items()
                if v is not None
            ]
        )

        return url

    def __get_safe_dsn(self, **connoptions):
        """
        Return a simplified dsn that won't leak the password if provided in the
        options, for logging purpose.
        """
        dsn = "{user}@{host}:{port}/{database}".format(**connoptions)
        return dsn

    def render_json(self, value):
        """
        Render the object as json response.
        """
        self.set_header("Content-Type", "application/json")
        self.write(to_json(value))

    @property
    def current_user(self):
        """
        Return the current_user if he is allowed to connect
        to his server of choice.
        """
        raw = self.get_str_cookie("user")
        if raw is not None:
            try:
                self.connect()
                return raw or "anonymous"
            except Exception:
                return None

    @property
    def current_server(self):
        """
        Return the server connected to if any
        """
        try:
            return self.get_secure_cookie("server").decode("utf-8")
        except AttributeError:
            return None

    @property
    def current_host(self):
        """
        Return the host connected to if any
        """
        server = self.current_server
        if server is None:
            return None
        connoptions = options.servers[server].copy()
        host = "localhost"
        if "host" in connoptions:
            host = connoptions["host"]
        return "%s" % (host)

    @property
    def current_port(self):
        """
        Return the port connected to if any
        """
        server = self.current_server
        if server is None:
            return None
        connoptions = options.servers[server].copy()
        port = "5432"
        if "port" in connoptions:
            port = connoptions["port"]
        return "%s" % (port)

    @property
    def current_connection(self):
        """
        Return the host and port connected to if any
        """
        server = self.current_server
        if server is None:
            return None
        connoptions = options.servers[server].copy()
        host = "localhost"
        port = "5432"
        if "host" in connoptions:
            host = connoptions["host"]
        if "port" in connoptions:
            port = connoptions["port"]
        return "%s:%s" % (host, port)

    @property
    def database(self):
        """Return the current database."""
        return None

    @property
    def server(self):
        """Return the current database."""
        return None

    @property
    def notify_allowed(self):
        """
        Returns whether the current user is allowed to use the NOTIFY-based
        communication with powa-collector
        """
        return False

    def get_powa_version(self, **kwargs):
        version = self.execute(
            """
            SELECT regexp_replace(extversion, '(dev|beta\d*)', '') AS version
            FROM pg_extension
            WHERE extname = 'powa'
            """,
            **kwargs,
        )

        if len(version) == 0:
            return None

        version = version[0]["version"]
        if version is None:
            return None
        return [int(part) for part in version.split(".")]

    def get_pg_version_num(self, srvid=None, **kwargs):
        try:
            return int(
                self.execute(
                    """
                SELECT setting
                FROM pg_settings
                WHERE name = 'server_version_num'
                """,
                    srvid=srvid,
                    **kwargs,
                )[0]["setting"]
            )
        except Exception:
            return None

    def get_databases(self, srvid):
        """
        Return the list of databases in this instance.
        """
        if self.current_user:
            if self._databases is None:
                self._databases = [
                    d["datname"]
                    for d in self.execute(
                        """
                    SELECT p.datname
                    FROM {powa}.powa_databases p
                    LEFT JOIN pg_database d ON p.oid = d.oid
                    WHERE COALESCE(datallowconn, true)
                    AND srvid = %(srvid)s
                    ORDER BY DATNAME
                    """,
                        params={"srvid": srvid},
                    )
                ]
            return self._databases

    def deparse_srvid(self, srvid):
        if srvid == "0":
            return self.current_connection
        else:
            return self.execute(
                """
                                SELECT COALESCE(alias,
                                                hostname || ':' || port)
                                       AS server
                                FROM {powa}.powa_servers
                                WHERE id = %(srvid)s
                                """,
                params={"srvid": int(srvid)},
            )[0]["server"]

    @property
    def servers(self, **kwargs):
        """
        Return the list of servers.
        """
        if self.current_user:
            if self._servers is None:
                self._servers = [
                    [s["id"], s["val"], s["alias"]]
                    for s in self.execute(
                        """
                    SELECT s.id, CASE WHEN s.id = 0 THEN
                        %(default)s
                    ELSE
                        s.hostname || ':' || s.port
                    END AS val,
                    s.alias
                    FROM {powa}.powa_servers s
                    ORDER BY hostname
                    """,
                        params={"default": self.current_connection},
                    )
                ]
            return self._servers

    def on_finish(self):
        for conn in self._connections.values():
            conn.close()

    def connect(
        self,
        srvid=None,
        server=None,
        user=None,
        password=None,
        database=None,
        remote_access=False,
        **kwargs,
    ):
        """
        Connect to a specific database.
        Parameters default values are taken from the cookies and the server
        configuration file.
        """
        if srvid is not None and srvid != "0":
            remote_access = True

        # Check for global connection restriction first
        if remote_access and not options["allow_ui_connection"]:
            raise Exception("UI connection globally not allowed.")

        conn_allowed = None
        server = server or self.get_str_cookie("server")
        user = user or self.get_str_cookie("user")
        password = password or self.get_str_cookie("password")
        if server not in options.servers:
            raise HTTPError(404, "Server %s not found." % server)

        connoptions = options.servers[server].copy()

        # Handle "query" parameter.  It should be a dict contain a single
        # client_encoding key.
        encoding_query = connoptions.pop("query", None)
        if encoding_query is not None:
            if not isinstance(encoding_query, dict):
                raise Exception(
                    'Invalid "query" parameter: %r, ' % encoding_query
                )

            for k in encoding_query:
                if k != "client_encoding":
                    raise Exception(
                        'Invalid "query" parameter: %r", '
                        'unexpected key "%s"' % (encoding_query, k)
                    )

        if srvid is not None and srvid != "0":
            tmp = self.connect()
            cur = tmp.cursor(cursor_factory=RealDictCursor)
            cur.execute(
                """
            SELECT hostname, port, username, password, dbname,
                allow_ui_connection
            FROM {powa}.powa_servers WHERE id = %(srvid)s
            """,
                {"srvid": srvid},
            )
            row = cur.fetchone()
            cur.close()

            connoptions["host"] = row["hostname"]
            connoptions["port"] = row["port"]
            connoptions["user"] = row["username"]
            connoptions["password"] = row["password"]
            connoptions["database"] = row["dbname"]
            conn_allowed = row["allow_ui_connection"]
        else:
            if "user" not in connoptions:
                connoptions["user"] = user
            if "password" not in connoptions:
                connoptions["password"] = password

        # If a non-powa connection is requested, check if we're allowed
        if remote_access:
            # authorization check for local connection has not been done yet
            if conn_allowed is None:
                tmp = self.connect(
                    remote_access=False,
                    server=server,
                    user=user,
                    password=password,
                )
                cur = tmp.cursor()
                cur.execute("""
                SELECT allow_ui_connection
                FROM {powa}.powa_servers WHERE id = 0
                """)
                row = cur.fetchone()
                cur.close()
                conn_allowed = row[0]

            if not conn_allowed:
                raise Exception("UI connection not allowed for this server.")

        if database is not None:
            connoptions["database"] = database

        url = self.__get_url(**connoptions)
        if url in self._connections:
            return self._connections.get(url)

        conn = psycopg2.connect(
            connection_factory=CustomConnection, **connoptions
        )
        conn.initialize(
            self.logger,
            srvid,
            self.__get_safe_dsn(**connoptions),
            encoding_query,
            self.application.settings["debug"],
        )

        # Get and cache all extensions schemas, in a dict with the extension
        # name as the key and the *quoted* schema as the value.
        cur = conn.cursor()
        cur.execute("""
            SELECT extname, quote_ident(nspname) AS nsp
            FROM pg_catalog.pg_extension e
            JOIN pg_catalog.pg_namespace n ON n.oid = e.extnamespace
        """)
        ext_nsps = {row[0]: row[1] for row in cur.fetchall()}
        cur.close()
        conn._nsps = ext_nsps

        # Cache the connection and return it
        self._connections[url] = conn
        return self._connections[url]

    def __get_extension_version(
        self, srvid, extname, database=None, remote_access=True
    ):
        """
        Returns a tuple with all digits of the version of the specific
        extension on the specific server and database, or None if the extension
        isn't install (or if we fail trying to retrieve the information).
        """
        remver = None

        # make sure we have a consistent type for the server id
        srvid = str(srvid)

        # Check for a cached version first.  Note that we do cache the lack of
        # extension (storing None), so we use an empty string to detect that no
        # caching happened yet.
        remver = self._ext_versions[srvid][database].get(extname, "")

        if remver != "":
            return remver

        # For remote server, check first if powa-collector reported a version
        # for that extension, but only for default database.
        if srvid != "0" and database is None:
            try:
                remver = self.execute(
                    """
                    SELECT version
                    FROM {powa}.powa_extension_config
                    WHERE srvid = %(srvid)s
                    AND extname = %(extname)s
                    """,
                    params={"srvid": srvid, "extname": extname},
                )[0]["version"]

            except Exception:
                return None
        else:
            # Otherwise, fall back to querying on the target database.
            try:
                remver = self.execute(
                    """
                    SELECT extversion
                    FROM pg_catalog.pg_extension
                    WHERE extname = %(extname)s
                    LIMIT 1
                    """,
                    srvid=srvid,
                    database=database,
                    params={"extname": extname},
                    remote_access=remote_access,
                )[0]["extversion"]
            except Exception:
                return None

        if remver is None:
            self._ext_versions[srvid][database][extname] = None
            return None

        # Clean up any extraneous characters
        remver = re.search(r"[0-9\.]*[0-9]", remver)

        if remver is None:
            self._ext_versions[srvid][database][extname] = None
            return None

        remver = tuple(map(int, remver.group(0).split(".")))
        self._ext_versions[srvid][database][extname] = remver

        return remver

    def has_extension(self, srvid, extname):
        """
        Returns whether the specific extensions is supposed to be installed, in
        any version, on the specific server for the default database.  This
        function is less expensive than has_extension_version for remote
        servers, as we're only checking for the metadata stored in the
        powa_functions table of the dedicated repository server instead of
        connecting to the remote server.  It also makes it possible to handle
        more widgets in the UI if the remotes servers are not accessible from
        the powa-web server.  This assumes that "module" is the name of the
        underlying extension.
        """
        if srvid == "0" or srvid == 0:
            # if local server, fallback to the full test, as it won't be more
            # expensive
            return self.has_extension_version(
                srvid, extname, "0", remote_access=False
            )
        else:
            try:
                # Look for at least an enabled snapshot function.  If a module
                # provides multiple snapshot functions and only a subset is
                # activated, let's assume that the extension is available.
                return self.execute(
                    """
                SELECT COUNT(*) != 0 AS res
                FROM {powa}.powa_functions
                WHERE srvid = %(srvid)s
                AND name = %(extname)s
                AND enabled
                """,
                    params={"srvid": srvid, "extname": extname},
                )[0]["res"]
            except Exception:
                return False

    def has_extension_version(
        self, srvid, extname, version, database=None, remote_access=True
    ):
        """
        Returns whether the version of the specific extension on the specific
        server and database is at least the given version.
        """
        if version is None:
            raise Exception("No version provided!")

        remver = self.__get_extension_version(
            srvid, extname, remote_access=remote_access
        )

        if remver is None:
            return False

        wanted = tuple(map(int, version.split(".")))

        return remver >= wanted

    def write_error(self, status_code, **kwargs):
        if status_code == 403:
            self._status_code = status_code
            self.clear_all_cookies()
            self.flash("Authentification failed", "alert")
            self.render("login.html", title="Login")
            return
        if status_code == 501:
            self._status_code = status_code
            self.render("xhr.html", content=kwargs["exc_info"][1].log_message)
            return
        super(BaseHandler, self).write_error(status_code, **kwargs)

    def execute(
        self,
        query,
        srvid=None,
        params=None,
        server=None,
        user=None,
        database=None,
        password=None,
        remote_access=False,
    ):
        """
        Execute a query against a database, with specific bind parameters.
        """
        if params is None:
            params = {}

        if "samples" not in params:
            params["samples"] = 100

        conn = self.connect(
            srvid, server, user, password, database, remote_access
        )

        cur = conn.cursor(cursor_factory=RealDictCursor)
        cur.execute("SAVEPOINT powa_web")
        try:
            cur.execute(query, params)

            # Fetch all results if any, and return them.
            if cur.rowcount > 0:
                rows = cur.fetchall()
            else:
                rows = []

            cur.execute("RELEASE powa_web")

        except Exception as e:
            cur.execute("ROLLBACK TO powa_web")
            raise e
        finally:
            cur.close()
        return rows

    def notify_collector(self, command, args=[], timeout=3):
        """
        Notify powa-collector and get its answer.
        """
        conn = self.connect()

        cur = conn.cursor()

        # we shouldn't listen on anything else than our own channel, but just
        # in case discard everything
        cur.execute("UNLISTEN *")

        random.seed()
        channel = "r%d" % random.randint(1, 99999)
        cur.execute("LISTEN %s" % channel)

        # some commands will contain user-provided strings, so we need to
        # properly escape the arguments.
        payload = "%s %s %s" % (command, channel, " ".join(args))
        cur.execute("NOTIFY powa_collector, %s", (payload,))

        cur.close()
        conn.commit()

        # wait for activity on the connection up to given timeout
        select.select([conn], [], [], timeout)

        # we shouldn't listen on anything else than our own channel, but just
        # in case discard everything
        cur = conn.cursor()
        cur.execute("UNLISTEN *")
        conn.commit()

        conn.poll()
        res = []
        while conn.notifies:
            notif = conn.notifies.pop(0)

            payload = notif.payload.split(" ")

            received_command = payload.pop(0)
            # we shouldn't received unexpected messages, but ignore them if any
            if received_command != command:
                continue

            status = payload.pop(0)
            payload = " ".join(payload)

            # we should get a single answer, but if multiple are received
            # append them and let the caller handle it.
            res.append({status: payload})

        return res

    def get_pickle_cookie(self, name):
        """
        Deserialize a cookie value using the pickle protocol.
        """
        value = self.get_secure_cookie(name)
        if value:
            try:
                return pickle.loads(value)
            except Exception:
                self.clear_all_cookies()

    def get_str_cookie(self, name, default=None):
        value = self.get_secure_cookie(name)
        if value is not None:
            return value.decode("utf8")
        return default

    def set_pickle_cookie(self, name, value):
        """
        Serialize a cookie value using the pickle protocol.
        """
        self.set_secure_cookie(name, pickle.dumps(value))

    flash = ui_methods.flash

    def to_json(self):
        return {
            "database": self.database,
            "currentPort": self.current_port,
            "currentServer": self.current_server,
            "currentUser": self.current_user,
            "currentConnection": self.current_connection,
            "notifyAllowed": self.notify_allowed,
            "server": self.server,
            "version": ui_methods.version(None),
            "year": ui_methods.year(None),
            "configUrl": self.reverse_url("RepositoryConfigOverview"),
            "logoUrl": self.static_url("img/favicon/favicon-32x32.png"),
            "homeUrl": self.reverse_url("Overview"),
        }


class AuthHandler(BaseHandler):
    """
    Base handler for urls needing authentifications.
    """

    @authenticated
    def prepare(self):
        super(AuthHandler, self).prepare()

    def to_json(self):
        return dict(
            **super(AuthHandler, self).to_json(),
            **{"logoutUrl": self.reverse_url("logout")},
        )



block_size = (
    "(SELECT cast(current_setting('block_size') AS numeric)"
    " AS block_size) AS bs"
)


def mulblock(col, alias=None, fn=None):
    alias = alias or col

    if fn is None:
        sql = "{col}"
    else:
        sql = "{fn}({col})"

    sql += " * block_size AS {alias}"

    return sql.format(col=col, alias=alias, fn=fn)


def total_measure_interval(col):
    sql = (
        "extract(epoch FROM "
        + " CASE WHEN min({col}) = '0 second' THEN '1 second'"
        + " ELSE min({col})"
        + "END)"
    )

    return sql.format(col=col)


def diff(var, alias=None):
    alias = alias or var
    return "max({var}) - min({var}) AS {alias}".format(var=var, alias=alias)


def diffblk(var, blksize=8192, alias=None):
    alias = alias or var
    return "(max({var}) - min({var})) * {blksize} AS {alias}".format(
        var=var, blksize=blksize, alias=alias
    )


def get_ts():
    return "extract(epoch FROM greatest(mesure_interval, '1 second'))"


def sum_per_sec(col, prefix=None, alias=None):
    alias = alias or col
    if prefix is not None:
        prefix = prefix + "."
    else:
        prefix = ""

    return "sum({prefix}{col}) / {ts} AS {alias}".format(
        prefix=prefix, col=col, ts=get_ts(), alias=alias
    )


def byte_per_sec(col, prefix=None, alias=None):
    alias = alias or col
    if prefix is not None:
        prefix = prefix + "."
    else:
        prefix = ""

    return "sum({prefix}{col}) * block_size / {ts} AS {alias}".format(
        prefix=prefix, col=col, ts=get_ts(), alias=alias
    )


def wps(col, do_sum=True):
    field = "sub." + col
    if do_sum:
        field = "sum(" + field + ")"

    return "({field} / {ts}) AS {col}".format(
        field=field, col=col, ts=get_ts()
    )


def to_epoch(col, prefix=None):
    if prefix is not None:
        qn = "{prefix}.{col}".format(prefix=prefix, col=col)
    else:
        qn = col

    return "extract(epoch FROM {qn}) AS {col}".format(qn=qn, col=col)


def total_read(prefix, noalias=False):
    if noalias:
        alias = ""
    else:
        alias = " AS total_blks_read"

    sql = (
        "sum({prefix}.shared_blks_read"
        + "+ {prefix}.local_blks_read"
        + "+ {prefix}.temp_blks_read"
        ") * block_size / {total_measure_interval}{alias}"
    )

    return sql.format(
        prefix=prefix,
        total_measure_interval=total_measure_interval("mesure_interval"),
        alias=alias,
    )


def total_hit(c):
    return (
        "sum(shared_blks_hit + local_blks_hit) * block_size /"
        + total_measure_interval("mesure_interval")
        + " AS total_blks_hit"
    )



"""
Functions to generate the queries used in the various grid components
"""

from powa.sql.utils import block_size, diff, diffblk


def powa_base_statdata_detailed_db():
    """
    Base for query used in the grids displaying info about pgss.

    This is based on the "detailed" version of the table, with queryid
    information.

    As the data is stored in 2 sets of tables (the coalesced records and the
    "current" records), we have to retrieve the query in multiple steps.

    To improve performance, the retrieval of the coalesced records is divided
    in 3 parts, which can greatly limit the number of rows that will need to be
    sorted.

    Those 3 parts are split in 3 subqueries.  For the left and right bounds
    (the smallest and largest timestamp in the given interval), we get the 2
    underlying records in the coalesced records (if any) and unnest them
    fully, filtering any records outside of the interval bound.  For any record
    in the coalesced records that are entirely inside the given interval,
    we rely on the pre-computed metadata (mins_in_range and maxs_in_range) and
    simply return that, which can greatly reduces the amount of rows to sort.

    For the "current" records, we simply return all the rows in the given
    interval.
    """
    base_query = """
  {powa}.powa_databases,
  LATERAL
  (
    -- Left bound: the search interval is a single timestamp, the smallest one
    -- of the search interval, and has to be inside the coalesce_range. We
    -- still need to unnest this one as we may have to remove some of the
    -- underlying records
    SELECT unnested.dbid, unnested.toplevel, unnested.userid, unnested.queryid,
      (unnested.records).*
    FROM (
      SELECT psh.dbid, psh.toplevel, psh.userid, psh.queryid,
        psh.coalesce_range, unnest(records) AS records
      FROM {powa}.powa_statements_history psh
      WHERE coalesce_range && tstzrange(%(from)s, %(from)s, '[]')
      AND psh.dbid = powa_databases.oid
      AND psh.srvid = %(server)s
    ) AS unnested
    WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- Right bound: the search interval is a single timestamp, the largest one
    -- of the search interval, and has to be inside the coalesce_range. We
    -- still need to unnest this one as we may have to remove some of the
    -- underlying records
    SELECT unnested.dbid, unnested.toplevel, unnested.userid, unnested.queryid,
      (unnested.records).*
    FROM (
      SELECT psh.dbid, psh.toplevel, psh.userid, psh.queryid,
        psh.coalesce_range, unnest(records) AS records
      FROM {powa}.powa_statements_history psh
      WHERE coalesce_range && tstzrange(%(to)s, %(to)s, '[]')
      AND psh.dbid = powa_databases.oid
      AND psh.srvid = %(server)s
    ) AS unnested
    WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- These entries have their coalesce_range ENTIRELY inside the search range
    -- so we don't need to unnest them.  We just retrieve the mins_in_range,
    -- maxs_in_range from the record, build an array of this and return it as
    -- if it was the full record
    SELECT unnested.dbid, unnested.toplevel, unnested.userid, unnested.queryid,
      (unnested.records).*
    FROM (
      SELECT psh.dbid, psh.toplevel, psh.userid, psh.queryid,
        psh.coalesce_range,
        unnest(ARRAY[mins_in_range,maxs_in_range]) AS records
      FROM {powa}.powa_statements_history psh
      WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
      AND psh.dbid = powa_databases.oid
      AND psh.srvid = %(server)s
    ) AS unnested
    WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- The "current" records are simply returned after filtering
    SELECT psc.dbid, psc.toplevel, psc.userid, psc.queryid,(psc.record).*
    FROM {powa}.powa_statements_history_current psc
    WHERE (record).ts <@ tstzrange(%(from)s,%(to)s,'[]')
    AND psc.dbid = powa_databases.oid
    AND psc.srvid = %(server)s
  ) h"""
    return base_query


def powa_base_statdata_db():
    """
    Query used in the grids displaying info about pgss.

    This is based on the db-aggregated version of the tables, without queryid
    information.

    This uses the same optimization as powa_base_statdata_detailed_db.
    """
    base_query = """(
 SELECT d.srvid, d.oid as dbid, h.*
 FROM
 {powa}.powa_databases d LEFT JOIN
 (
   SELECT srvid, dbid
   FROM {powa}.powa_statements_history_db dbh
   WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
   AND dbh.srvid = %(server)s
   GROUP BY srvid, dbid
 ) ranges ON d.oid = ranges.dbid AND d.srvid = ranges.srvid,
 LATERAL (
   -- Left bound: the search interval is a single timestamp, the smallest one
   -- of the search interval, and has to be inside the coalesce_range. We
   -- still need to unnest this one as we may have to remove some of the
   -- underlying records
   SELECT (unnested1.records).*
   FROM (
     SELECT dbh.coalesce_range, unnest(records) AS records
     FROM {powa}.powa_statements_history_db dbh
     WHERE coalesce_range && tstzrange(%(from)s, %(from)s, '[]')
     AND dbh.dbid = ranges.dbid
     AND dbh.srvid = %(server)s
   ) AS unnested1
   WHERE (unnested1.records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

   -- Right bound: the search interval is a single timestamp, the largest one
   -- of the search interval, and has to be inside the coalesce_range. We
   -- still need to unnest this one as we may have to remove some of the
   -- underlying records
   UNION ALL
   SELECT (unnested2.records).*
   FROM (
     SELECT dbh.coalesce_range, unnest(records) AS records
     FROM {powa}.powa_statements_history_db dbh
     WHERE coalesce_range && tstzrange(%(to)s, %(to)s, '[]')
     AND dbh.dbid = ranges.dbid
     AND dbh.srvid = %(server)s
   ) AS unnested2
   WHERE  (unnested2.records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

   UNION ALL

   -- These entries have their coalesce_range ENTIRELY inside the search range
   -- so we don't need to unnest them.  We just retrieve the mins_in_range,
   -- maxs_in_range from the record, build an array of this and return it as
   -- if it was the full record
   SELECT (unnested1.records).*
   FROM (
     SELECT dbh.coalesce_range,
       unnest(ARRAY[mins_in_range,maxs_in_range]) AS records
     FROM {powa}.powa_statements_history_db dbh
     WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
     AND dbh.dbid = ranges.dbid
     AND dbh.srvid = %(server)s
   ) AS unnested1
   WHERE (unnested1.records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

   UNION ALL

   -- The "current" records are simply returned after filtering
   SELECT (dbc.record).*
   FROM {powa}.powa_statements_history_current_db dbc
   WHERE  (dbc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
   AND dbc.dbid = d.oid
   AND dbc.srvid = d.srvid
   AND dbc.srvid = %(server)s
    ) AS h
) AS db_history
    """
    return base_query


def powa_base_bgwriter():
    base_query = """(
 SELECT h.*
 FROM
 (
   SELECT srvid
   FROM {powa}.powa_stat_bgwriter_history bgwh
   WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
   AND dbh.srvid = %(server)s
   GROUP BY srvid
 ) ranges
 LATERAL (
   SELECT (unnested1.records).*
   FROM (
     SELECT bgwh.coalesce_range, unnest(records) AS records
     FROM {powa}.powa_stat_bgwriter_history bgwh
     WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
     AND bgwh.srvid = %(server)s
   ) AS unnested1
   WHERE  (unnested1.records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
   UNION ALL
   SELECT (bgwc.record).*
   FROM {powa}.powa_stat_bgwriter_history_current bgwc
   WHERE  (dbc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
   AND dbc.srvid = d.srvid
   AND dbc.srvid = %(server)s
    ) AS h
) AS bgw_history
    """
    return base_query


def powa_base_io():
    """
    Query used in the grids displaying info about pg_stat_io.

    This uses the same optimization as powa_base_statdata_detailed_db.
    """
    base_query = """(
 SELECT *
 FROM
 (
   SELECT srvid
   FROM {powa}.powa_stat_io_history ioh
   WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
   AND ioh.srvid = %(server)s
   GROUP BY srvid
 ) ranges,
 LATERAL (
   -- Left bound: the search interval is a single timestamp, the smallest one
   -- of the search interval, and has to be inside the coalesce_range. We
   -- still need to unnest this one as we may have to remove some of the
   -- underlying records
   SELECT backend_type, object, context, (unnested.records).*
   FROM (
     SELECT ioh.backend_type, ioh.object, ioh.context,
       ioh.coalesce_range, unnest(records) AS records
     FROM {powa}.powa_stat_io_history ioh
     WHERE coalesce_range && tstzrange(%(from)s, %(from)s, '[]')
     AND ioh.srvid = %(server)s
   ) AS unnested
   WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

   UNION ALL

   -- Right bound: the search interval is a single timestamp, the largest one
   -- of the search interval, and has to be inside the coalesce_range. We
   -- still need to unnest this one as we may have to remove some of the
   -- underlying records
   SELECT backend_type, object, context, (unnested.records).*
   FROM (
     SELECT ioh.backend_type, ioh.object, ioh.context,
       ioh.coalesce_range, unnest(records) AS records
     FROM {powa}.powa_stat_io_history ioh
     WHERE coalesce_range && tstzrange(%(to)s, %(to)s, '[]')
     AND ioh.srvid = %(server)s
   ) AS unnested
   WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

   UNION ALL

   -- These entries have their coalesce_range ENTIRELY inside the search range
   -- so we don't need to unnest them.  We just retrieve the mins_in_range,
   -- maxs_in_range from the record, build an array of this and return it as
   -- if it was the full record
   SELECT backend_type, object, context, (unnested.records).*
   FROM (
     SELECT ioh.backend_type, ioh.object, ioh.context,
       ioh.coalesce_range,
       unnest(ARRAY[mins_in_range,maxs_in_range]) AS records
     FROM {powa}.powa_stat_io_history ioh
     WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
     AND ioh.srvid = %(server)s
   ) AS unnested
   WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
 ) AS h
 UNION ALL
 SELECT srvid, backend_type, object, context, (ioc.record).*
 FROM {powa}.powa_stat_io_history_current ioc
 WHERE (ioc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
 AND ioc.srvid = %(server)s
) AS io_history
    """
    return base_query


def powa_getiodata(qual=None):
    """
    Query used in the grid displaying info about pg_stat_io.
    """
    base_query = powa_base_io()

    if qual is not None:
        qual = " WHERE %s" % qual
    else:
        qual = ""

    base_cols = [
        "srvid",
        "backend_type",
        "object",
        "context",
    ]

    cols = [
        diffblk("reads", "op_bytes"),
        diff("read_time"),
        diffblk("writes", "op_bytes"),
        diff("write_time"),
        diffblk("writebacks", "op_bytes"),
        diff("writeback_time"),
        diffblk("extends", "op_bytes"),
        diff("extend_time"),
        diffblk("hits", "op_bytes"),
        diffblk("evictions", "op_bytes"),
        diffblk("reuses", "op_bytes"),
        diffblk("fsyncs", "op_bytes"),
        diff("fsync_time"),
    ]

    return """SELECT {base_cols}, {cols}
    FROM {base_query}
    {qual}
    GROUP BY srvid, {base_cols}, op_bytes""".format(
        cols=", ".join(cols),
        base_cols=", ".join(base_cols),
        base_query=base_query,
        qual=qual,
    )


def powa_base_slru():
    """
    Query used in the grids displaying info about pg_stat_slru.

    This uses the same optimization as powa_base_statdata_detailed_db.
    """
    base_query = """(
 SELECT *
 FROM
 (
   SELECT srvid
   FROM {powa}.powa_stat_slru_history slru
   WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
   AND slru.srvid = %(server)s
   GROUP BY srvid
 ) ranges,
 LATERAL (
   -- Left bound: the search interval is a single timestamp, the smallest one
   -- of the search interval, and has to be inside the coalesce_range. We
   -- still need to unnest this one as we may have to remove some of the
   -- underlying records
   SELECT name, (unnested.records).*
   FROM (
     SELECT slru.name, slru.coalesce_range, unnest(records) AS records
     FROM {powa}.powa_stat_slru_history slru
     WHERE coalesce_range && tstzrange(%(from)s, %(from)s, '[]')
     AND slru.srvid = %(server)s
   ) AS unnested
   WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

   UNION ALL

   -- Right bound: the search interval is a single timestamp, the largest one
   -- of the search interval, and has to be inside the coalesce_range. We
   -- still need to unnest this one as we may have to remove some of the
   -- underlying records
   SELECT name, (unnested.records).*
   FROM (
     SELECT slru.name, slru.coalesce_range, unnest(records) AS records
     FROM {powa}.powa_stat_slru_history slru
     WHERE coalesce_range && tstzrange(%(to)s, %(to)s, '[]')
     AND slru.srvid = %(server)s
   ) AS unnested
   WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

   UNION ALL

   -- These entries have their coalesce_range ENTIRELY inside the search range
   -- so we don't need to unnest them.  We just retrieve the mins_in_range,
   -- maxs_in_range from the record, build an array of this and return it as
   -- if it was the full record
   SELECT name, (unnested.records).*
   FROM (
     SELECT slru.name, slru.coalesce_range,
       unnest(ARRAY[mins_in_range,maxs_in_range]) AS records
     FROM {powa}.powa_stat_slru_history slru
     WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
     AND slru.srvid = %(server)s
   ) AS unnested
   WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
 ) AS h
 UNION ALL
 SELECT srvid, name, (ioc.record).*
 FROM {powa}.powa_stat_slru_history_current ioc
 WHERE  (ioc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
 AND ioc.srvid = %(server)s
) AS io_history
    """
    return base_query


def powa_getslrudata(qual=None):
    """
    Query used in the grid displaying info about pg_stat_slru.
    """
    base_query = powa_base_slru()

    if qual is not None:
        qual = " WHERE %s" % qual
    else:
        qual = ""

    base_cols = [
        "srvid",
        "name",
    ]

    cols = [
        diffblk("blks_zeroed", "block_size"),
        diffblk("blks_hit", "block_size"),
        diffblk("blks_read", "block_size"),
        diffblk("blks_written", "block_size"),
        diffblk("blks_exists", "block_size"),
        diff("flushes"),
        diff("truncates"),
    ]

    return """SELECT {base_cols}, {cols}
    FROM {base_query}
    CROSS JOIN {block_size}
    {qual}
    GROUP BY {base_cols}, block_size""".format(
        cols=", ".join(cols),
        base_cols=", ".join(base_cols),
        base_query=base_query,
        block_size=block_size,
        qual=qual,
    )


def get_diffs_forstatdata():
    return [
        diff("calls"),
        diff("total_plan_time", "plantime"),
        diff("total_exec_time", "runtime"),
        diff("shared_blks_read"),
        diff("shared_blks_hit"),
        diff("shared_blks_dirtied"),
        diff("shared_blks_written"),
        diff("temp_blks_read"),
        diff("temp_blks_written"),
        diff("shared_blk_read_time"),
        diff("shared_blk_write_time"),
        diff("local_blk_read_time"),
        diff("local_blk_write_time"),
        diff("temp_blk_read_time"),
        diff("temp_blk_write_time"),
        diff("wal_records"),
        diff("wal_fpi"),
        diff("wal_bytes"),
        diff("jit_functions"),
        diff("jit_generation_time"),
        diff("jit_inlining_count"),
        diff("jit_inlining_time"),
        diff("jit_optimization_count"),
        diff("jit_optimization_time"),
        diff("jit_emission_count"),
        diff("jit_emission_time"),
        diff("jit_deform_count"),
        diff("jit_deform_time"),
    ]


def powa_getstatdata_detailed_db(srvid="%(server)s", predicates=[]):
    """
    Query used in the grids displaying info about pgss.

    This is based on the "detailed" version of the tables, with queryid
    information.

    predicates is an optional array of plain-text predicates.
    """
    base_query = powa_base_statdata_detailed_db()
    diffs = get_diffs_forstatdata()

    where = " AND ".join(predicates)
    if where != "":
        where = " AND " + where

    cols = [
        "srvid",
        "queryid",
        "dbid",
        "toplevel",
        "userid",
        "datname",
    ] + diffs

    return """SELECT {cols}
    FROM {base_query}
    WHERE srvid = {srvid}
    {where}
    GROUP BY srvid, queryid, dbid, toplevel, userid, datname
    HAVING max(calls) - min(calls) > 0""".format(
        cols=", ".join(cols), base_query=base_query, srvid=srvid, where=where
    )


def powa_getstatdata_db(srvid):
    """
    Query used in the grids displaying info about pgss.

    This is based on the db-aggregated version of the tables, without queryid
    information.
    """
    base_query = powa_base_statdata_db()
    diffs = get_diffs_forstatdata()

    cols = ["srvid", "dbid"] + diffs

    return """SELECT {cols}
    FROM {base_query}
    WHERE srvid = {srvid}
    GROUP BY srvid, dbid
    HAVING max(calls) - min(calls) > 0""".format(
        cols=", ".join(cols), base_query=base_query, srvid=srvid
    )


def powa_base_waitdata_detailed_db():
    """
    Query used in the grids displaying info about pg_wait_sampling.

    This is based on the "detailed" version of the tables, with queryid
    information.

    This uses the same optimization as powa_base_statdata_detailed_db.
    """
    base_query = """
  {powa}.powa_databases,
  LATERAL
  (
    -- Left bound: the search interval is a single timestamp, the smallest one
    -- of the search interval, and has to be inside the coalesce_range. We
    -- still need to unnest this one as we may have to remove some of the
    -- underlying records
    SELECT unnested.dbid, unnested.queryid,
      unnested.event_type, unnested.event, (unnested.records).*
    FROM (
      SELECT wsh.dbid, wsh.queryid, wsh.event_type, wsh.event,
        wsh.coalesce_range, unnest(records) AS records
      FROM {powa}.powa_wait_sampling_history wsh
      WHERE coalesce_range && tstzrange(%(from)s, %(from)s, '[]')
      AND wsh.dbid = powa_databases.oid
      -- we can't simply join powa_statements as there's no userid in
      -- powa_wait_sampling_* tables
      AND wsh.srvid = %(server)s
    ) AS unnested
    WHERE  (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- Right bound: the search interval is a single timestamp, the largest one
    -- of the search interval, and has to be inside the coalesce_range. We
    -- still need to unnest this one as we may have to remove some of the
    -- underlying records
    SELECT unnested.dbid, unnested.queryid,
      unnested.event_type, unnested.event, (unnested.records).*
    FROM (
      SELECT wsh.dbid, wsh.queryid, wsh.event_type, wsh.event,
        wsh.coalesce_range, unnest(records) AS records
      FROM {powa}.powa_wait_sampling_history wsh
      WHERE coalesce_range && tstzrange(%(to)s, %(to)s, '[]')
      AND wsh.dbid = powa_databases.oid
      -- we can't simply join powa_statements as there's no userid in
      -- powa_wait_sampling_* tables
      AND wsh.srvid = %(server)s
    ) AS unnested
    WHERE  (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- These entries have their coalesce_range ENTIRELY inside the search range
    -- so we don't need to unnest them.  We just retrieve the mins_in_range,
    -- maxs_in_range from the record, build an array of this and return it as
    -- if it was the full record
    SELECT unnested.dbid, unnested.queryid,
      unnested.event_type, unnested.event, (unnested.records).*
    FROM (
      SELECT wsh.dbid, wsh.queryid, wsh.event_type, wsh.event,
        wsh.coalesce_range,
        unnest(ARRAY[mins_in_range,maxs_in_range]) AS records
      FROM {powa}.powa_wait_sampling_history wsh
      WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
      AND wsh.dbid = powa_databases.oid
      -- we can't simply join powa_statements as there's no userid in
      -- powa_wait_sampling_* tables
      AND wsh.srvid = %(server)s
    ) AS unnested
    WHERE  (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- The "current" records are simply returned after filtering
    SELECT wsc.dbid, wsc.queryid, wsc.event_type, wsc.event, (wsc.record).*
    FROM {powa}.powa_wait_sampling_history_current wsc
    WHERE (record).ts <@ tstzrange(%(from)s,%(to)s,'[]')
    AND wsc.dbid = powa_databases.oid
    -- we can't simply join powa_statements as there's no userid in
    -- powa_wait_sampling_* tables
    AND wsc.srvid = %(server)s
  ) h
  WHERE powa_databases.srvid = %(server)s
"""
    return base_query


def powa_base_waitdata_db():
    """
    Query used in the grids displaying info about pg_wait_sampling.

    This is based on the db-aggregated version of the tables, without queryid
    information.

    This uses the same optimization as powa_base_statdata_detailed_db.
    """
    base_query = """(
  SELECT powa_databases.srvid, powa_databases.oid as dbid,
    powa_databases.datname, h.*
  FROM
  {powa}.powa_databases LEFT JOIN
  (
    SELECT dbid
    FROM {powa}.powa_wait_sampling_history_db wsh
    WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
    AND wsh.srvid = %(server)s
    GROUP BY dbid
  ) ranges ON powa_databases.oid = ranges.dbid,
  LATERAL (
    -- Left bound: the search interval is a single timestamp, the smallest one
    -- of the search interval, and has to be inside the coalesce_range. We
    -- still need to unnest this one as we may have to remove some of the
    -- underlying records
    SELECT event_type, event, (unnested1.records).*
    FROM (
      SELECT wsh.event_type, wsh.event, unnest(records) AS records
      FROM {powa}.powa_wait_sampling_history_db wsh
      WHERE coalesce_range && tstzrange(%(from)s, %(from)s, '[]')
      AND wsh.dbid = ranges.dbid
      AND wsh.srvid = %(server)s
    ) AS unnested1
    WHERE (unnested1.records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- Right bound: the search interval is a single timestamp, the largest one
    -- of the search interval, and has to be inside the coalesce_range. We
    -- still need to unnest this one as we may have to remove some of the
    -- underlying records
    SELECT event_type, event, (unnested1.records).*
    FROM (
      SELECT wsh.event_type, wsh.event, unnest(records) AS records
      FROM {powa}.powa_wait_sampling_history_db wsh
      WHERE coalesce_range && tstzrange(%(to)s, %(to)s, '[]')
      AND wsh.dbid = ranges.dbid
      AND wsh.srvid = %(server)s
    ) AS unnested1
    WHERE (unnested1.records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- These entries have their coalesce_range ENTIRELY inside the search range
    -- so we don't need to unnest them.  We just retrieve the mins_in_range,
    -- maxs_in_range from the record, build an array of this and return it as
    -- if it was the full record
    SELECT event_type, event, (unnested1.records).*
    FROM (
      SELECT wsh.event_type, wsh.event,
        unnest(ARRAY[mins_in_range,maxs_in_range]) AS records
      FROM {powa}.powa_wait_sampling_history_db wsh
      WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
      AND wsh.dbid = ranges.dbid
      AND wsh.srvid = %(server)s
    ) AS unnested1
    WHERE (unnested1.records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- The "current" records are simply returned after filtering
    SELECT event_type, event, (wsc.record).*
    FROM {powa}.powa_wait_sampling_history_current_db wsc
    WHERE (wsc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
    AND wsc.dbid = powa_databases.oid
    AND wsc.srvid = %(server)s
  ) AS h
  WHERE powa_databases.srvid = %(server)s
) AS ws_history
    """
    return base_query


def powa_getwaitdata_detailed_db():
    """
    Query used in the grids displaying info about pg_wait_sampling.

    This is based on the "detailed" version of the tables, with queryid
    information.
    """
    base_query = powa_base_waitdata_detailed_db()
    return """SELECT srvid, queryid, dbid, datname, event_type, event,
    {count}
    FROM {base_query}
    GROUP BY srvid, queryid, dbid, datname, event_type, event
    HAVING max(count) - min(count) > 0""".format(
        count=diff("count"), base_query=base_query
    )


def powa_getwaitdata_db():
    """
    Query used in the grids displaying info about pg_wait_sampling.

    This is based on the db-aggregated version of the tables, without queryid
    information.
    """
    base_query = powa_base_waitdata_db()

    return """SELECT srvid, dbid, event_type, event, {count}
    FROM {base_query}
    GROUP BY srvid, dbid, event_type, event
    HAVING max(count) - min(count) > 0""".format(
        count=diff("count"), base_query=base_query
    )


def powa_base_userfuncdata_detailed_db():
    """
    Base for query used in the grids displaying info about
    pg_stat_user_functions.

    This is based on the "detailed" version of the tables, with per-function
    information.

    This uses the same optimization as powa_base_statdata_detailed_db.
    """
    base_query = """
  {powa}.powa_databases d
  JOIN LATERAL
  (
    -- Left bound: the search interval is a single timestamp, the smallest one
    -- of the search interval, and has to be inside the coalesce_range. We
    -- still need to unnest this one as we may have to remove some of the
    -- underlying records
    SELECT unnested.dbid, unnested.funcid,
      (unnested.records).*
    FROM (
      SELECT ufh.dbid, ufh.funcid,
        ufh.coalesce_range, unnest(records) AS records
      FROM {powa}.powa_user_functions_history ufh
      WHERE coalesce_range && tstzrange(%(from)s, %(from)s, '[]')
      AND ufh.dbid = d.oid
      AND ufh.srvid = %(server)s
    ) AS unnested
    WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- Right bound: the search interval is a single timestamp, the largest one
    -- of the search interval, and has to be inside the coalesce_range. We
    -- still need to unnest this one as we may have to remove some of the
    -- underlying records
    SELECT unnested.dbid, unnested.funcid,
      (unnested.records).*
    FROM (
      SELECT ufh.dbid, ufh.funcid,
        ufh.coalesce_range, unnest(records) AS records
      FROM {powa}.powa_user_functions_history ufh
      WHERE coalesce_range && tstzrange(%(to)s, %(to)s, '[]')
      AND ufh.dbid = d.oid
      AND ufh.srvid = %(server)s
    ) AS unnested
    WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- These entries have their coalesce_range ENTIRELY inside the search range
    -- so we don't need to unnest them.  We just retrieve the mins_in_range,
    -- maxs_in_range from the record, build an array of this and return it as
    -- if it was the full record
    SELECT unnested.dbid, unnested.funcid,
      (unnested.records).*
    FROM (
      SELECT ufh.dbid, ufh.funcid,
        ufh.coalesce_range,
        unnest(ARRAY[mins_in_range,maxs_in_range]) AS records
      FROM {powa}.powa_user_functions_history ufh
      WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
      AND ufh.dbid = d.oid
      AND ufh.srvid = %(server)s
    ) AS unnested
    WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- The "current" records are simply returned after filtering
    SELECT ufc.dbid, ufc.funcid, (ufc.record).*
    FROM {powa}.powa_user_functions_history_current ufc
    WHERE (record).ts <@ tstzrange(%(from)s,%(to)s,'[]')
    AND ufc.dbid = d.oid
    AND ufc.srvid = %(server)s
  ) h ON d.srvid = %(server)s
  LEFT JOIN {powa}.powa_catalog_proc pcp
    ON pcp.dbid = h.dbid AND pcp.oid = h.funcid AND pcp.srvid = d.srvid
        AND pcp.srvid = %(server)s
  LEFT JOIN {powa}.powa_catalog_language pcl
    ON pcl.srvid = pcp.srvid AND pcl.dbid = pcp.dbid AND pcl.oid = pcp.prolang"""
    return base_query


def powa_base_userfuncdata_db():
    """
    Base for query used in the grids displaying info about
    pg_stat_user_functions.

    This is based on the db-aggregated version of the tables, without
    per-function information.

    This uses the same optimization as powa_base_statdata_detailed_db.
    """
    base_query = """
  {powa}.powa_databases d,
  LATERAL
  (
    -- Left bound: the search interval is a single timestamp, the smallest one
    -- of the search interval, and has to be inside the coalesce_range. We
    -- still need to unnest this one as we may have to remove some of the
    -- underlying records
    SELECT unnested.dbid,
      (unnested.records).*
    FROM (
      SELECT ufhd.dbid,
        ufhd.coalesce_range, unnest(records) AS records
      FROM {powa}.powa_user_functions_history_db ufhd
      WHERE coalesce_range && tstzrange(%(from)s, %(from)s, '[]')
      AND ufhd.dbid = d.oid
      AND ufhd.srvid = %(server)s
    ) AS unnested
    WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- Right bound: the search interval is a single timestamp, the largest one
    -- of the search interval, and has to be inside the coalesce_range. We
    -- still need to unnest this one as we may have to remove some of the
    -- underlying records
    SELECT unnested.dbid,
      (unnested.records).*
    FROM (
      SELECT ufhd.dbid,
        ufhd.coalesce_range, unnest(records) AS records
      FROM {powa}.powa_user_functions_history_db ufhd
      WHERE coalesce_range && tstzrange(%(to)s, %(to)s, '[]')
      AND ufhd.dbid = d.oid
      AND ufhd.srvid = %(server)s
    ) AS unnested
    WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- These entries have their coalesce_range ENTIRELY inside the search range
    -- so we don't need to unnest them.  We just retrieve the mins_in_range,
    -- maxs_in_range from the record, build an array of this and return it as
    -- if it was the full record
    SELECT unnested.dbid,
      (unnested.records).*
    FROM (
      SELECT ufhd.dbid,
        ufhd.coalesce_range,
        unnest(ARRAY[mins_in_range,maxs_in_range]) AS records
      FROM {powa}.powa_user_functions_history_db ufhd
      WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
      AND ufhd.dbid = d.oid
      AND ufhd.srvid = %(server)s
    ) AS unnested
    WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')

    UNION ALL

    -- The "current" records are simply returned after filtering
    SELECT ufcd.dbid, (ufcd.record).*
    FROM {powa}.powa_user_functions_history_current_db ufcd
    WHERE (record).ts <@ tstzrange(%(from)s,%(to)s,'[]')
    AND ufcd.dbid = d.oid
    AND ufcd.srvid = %(server)s
  ) h
  LEFT JOIN {powa}.powa_catalog_proc pcp
    ON pcp.dbid = h.dbid AND pcp.srvid = %(server)s
  WHERE d.srvid = pcp.srvid"""
    return base_query


def powa_getuserfuncdata_detailed_db(funcid=None):
    """
    Query used in the grids displaying info about pg_stat_user_functions.

    This is based on the "detailed" version of the tables, with per-function
    information.
    """
    base_query = powa_base_userfuncdata_detailed_db()

    cols = [
        "d.srvid",
        "h.dbid",
        "d.datname",
        "funcid",
        "coalesce(regprocedure, '<' || funcid || '>') AS func_name",
        "lanname",
        diff("calls"),
        diff("total_time"),
        diff("self_time"),
    ]
    if funcid:
        cols.extend(["prosrc", "last_refresh"])

    groupby = [
        "d.srvid",
        "h.dbid",
        "d.datname",
        "funcid",
        "regprocedure",
        "lanname",
    ]
    if funcid:
        groupby.extend(["prosrc", "last_refresh"])

    if funcid:
        join_db = """INNER JOIN {powa}.powa_databases pd
            ON pd.srvid = d.srvid AND pd.oid = h.dbid"""
        and_funcid = "AND funcid = {funcid}".format(funcid=funcid)
    else:
        join_db = ""
        and_funcid = ""

    return """SELECT {cols}
    FROM {base_query}
    {join_db}
    WHERE d.datname = %(database)s
    {and_funcid}
    GROUP BY {groupby}
    HAVING max(calls) - min(calls) > 0""".format(
        cols=", ".join(cols),
        base_query=base_query,
        join_db=join_db,
        and_funcid=and_funcid,
        groupby=", ".join(groupby),
    )


def powa_getuserfuncdata_db():
    """
    Query used in the grids displaying info about pg_stat_user_functions.

    This is based on the db-aggregated version of the tables, without
    per-function information.
    """
    base_query = powa_base_userfuncdata_db()

    cols = [
        "d.srvid",
        "d.datname",
        "h.dbid",
        diff("calls"),
        diff("total_time"),
        diff("self_time"),
    ]

    return """SELECT {cols}
    FROM {base_query}
    GROUP BY d.srvid, d.datname, h.dbid, d.oid
    HAVING max(calls) - min(calls) > 0""".format(
        cols=", ".join(cols), base_query=base_query
    )



"""
Dashboard for the by-function page.
"""

from powa.dashboards import ContentWidget, Dashboard, DashboardPage
from powa.database import DatabaseOverview
from powa.sql.views_grid import powa_getuserfuncdata_detailed_db


class FunctionDetail(ContentWidget):
    """
    Detail widget showing summarized information for the function.
    """

    title = "Function Detail"
    data_url = r"/server/(\d+)/metrics/database/([^\/]+)/function/(\d+)/detail"

    def get(self, server, database, function):
        stmt = powa_getuserfuncdata_detailed_db("%(funcid)s")

        value = self.execute(
            stmt,
            params={
                "server": server,
                "database": database,
                "funcid": function,
                "from": self.get_argument("from"),
                "to": self.get_argument("to"),
            },
        )
        if value is None or len(value) < 1:
            self.render_json(None)
            return
        self.render_json(value[0])


class FunctionOverview(DashboardPage):
    """
    Dashboard page for a function.
    """

    base_url = r"/server/(\d+)/database/([^\/]+)/function/(\d+)/overview"
    params = ["server", "database", "function"]
    datasources = [FunctionDetail]
    parent = DatabaseOverview
    title = "Function overview"

    def dashboard(self):
        # This COULD be initialized in the constructor, but tornado < 3 doesn't
        # call it
        if getattr(self, "_dashboard", None) is not None:
            return self._dashboard

        self._dashboard = Dashboard(
            "Function %(function)s on database %(database)s",
            [[FunctionDetail]],
        )

        return self._dashboard



#!/usr/bin/env python
import logging
import tornado.ioloop
from powa import make_app
from tornado.options import options

if __name__ == "__main__":
    application = make_app(debug=True, gzip=True, compress_response=True)
    application.listen(options.port, address=options.address)
    logger = logging.getLogger("tornado.application")
    logger.info(
        "Starting powa-web on http://%s:%s%s",
        options.address,
        options.port,
        options.url_prefix,
    )
    tornado.ioloop.IOLoop.instance().start()



"""
Dashboard for the configuration summary page.
"""

from __future__ import absolute_import

import json
from powa.collector import CollectorServerDetail
from powa.dashboards import (
    ContentWidget,
    Dashboard,
    DashboardPage,
    Grid,
    MetricDef,
    MetricGroupDef,
)
from powa.sql.views import get_config_changes


def get_pgts_query(handler, restrict_database=False):
    # avoid pg errors if the extension hasn't been installed, or is not
    # installed in version 2.0.0 or more.  We check for local installation
    # only, as query will look in local table.  If the extension isn't setup
    # remotely, the check will be pretty quick.
    pgts = handler.has_extension_version(
        "0", "pg_track_settings", "2.0.0", remote_access=False
    )
    if not pgts:
        return None

    return get_config_changes(restrict_database)


class ConfigChangesGlobal(MetricGroupDef):
    name = "Config Changes"
    data_url = r"/server/(\d+)/config_changes"
    xaxis = "ts"
    axis_type = "category"
    setting = MetricDef(label="Name", type="string")
    previous = MetricDef(label="Previous", type="string")
    new = MetricDef(label="New", type="string")
    params = ["server"]

    @property
    def query(self):
        return get_pgts_query(self)


class ConfigChangesDatabase(MetricGroupDef):
    name = "Config Changes"
    data_url = r"/server/(\d+)/database/([^\/]+)/config_changes"
    xaxis = "ts"
    axis_type = "category"
    setting = MetricDef(label="Name", type="string")
    previous = MetricDef(label="Previous", type="string")
    new = MetricDef(label="New", type="string")
    params = ["server", "database"]

    @property
    def query(self):
        return get_pgts_query(self, True)


class ConfigChangesQuery(MetricGroupDef):
    name = "Config Changes"
    data_url = r"/server/(\d+)/database/([^\/]+)/query/(-?\d+)/config_changes"
    xaxis = "ts"
    axis_type = "category"
    setting = MetricDef(label="Name", type="string")
    previous = MetricDef(label="Previous", type="string")
    new = MetricDef(label="New", type="string")
    params = ["server", "database", "query"]

    @property
    def query(self):
        return get_pgts_query(self, True)


class ServersErrors(ContentWidget):
    """
    Detail widget showing the various errors related to servers logged
    """

    title = "Errors"
    data_url = r"/config/errors"

    def get(self):
        sql = """SELECT srvid,
            CASE WHEN id = 0 THEN
               '<local>'
            ELSE
               username || '@' || hostname || ':' || port || '/' || dbname
            END AS server_alias,
            errors
            FROM {powa}.powa_servers s
            JOIN {powa}.powa_snapshot_metas m ON m.srvid = s.id
            WHERE errors IS NOT NULL
            ORDER BY 1
        """

        rows = self.execute(sql)

        self.render_json(rows)


class AllCollectorsDetail(ContentWidget):
    """
    Detail widget showing summarized information for the background worker and
    the remote collector daemon.
    """

    title = "Collector Detail"
    data_url = r"/config/allcollectors"

    def get(self):
        sql = """SELECT id,
            CASE WHEN id = 'collector'
                THEN 'Remote collector'
                  || coalesce(
                    substring(application_name,
                        length('PoWA collector - main thread ')
                    ), '')
                ELSE 'Backgound worker'
            END AS powa_kind,
            date_trunc('second', backend_start) as start,
            datname, usename,
            coalesce(host(client_addr), '<local>') AS client_addr,
            count(datname) OVER () AS nb_found,
            (
                SELECT count(*) > 0
                FROM pg_stat_activity
                WHERE query = '<insufficient privilege>'
            ) AS not_authorized
            FROM (
                SELECT 'bgworker' AS id, 'PoWA - %%' AS val,
                    'powa' AS backend_type
                UNION ALL
                SELECT 'collector', 'PoWA collector - main thread (%%',
                    'client backend' AS backend_type
            ) n
            LEFT JOIN pg_stat_activity a ON
                a.application_name LIKE n.val
                AND a.backend_type = n.backend_type
            ORDER BY 1"""

        rows = self.execute(sql)

        self.logger.warn("%r", rows[0])
        if rows[0]["not_authorized"] is True:
            collector = None
        if rows[0]["nb_found"] == 0:
            collector = []
        else:
            collector = rows
        self.render_json(collector)


class PowaServersMetricGroup(MetricGroupDef):
    """
    Metric group for the servers list grid.
    """

    name = "powa_servers"
    xaxis = "server_alias"
    data_url = r"/config/powa_servers/"
    axis_type = "category"
    hostname = MetricDef(label="Hostname", type="string")
    port = MetricDef(label="Port", type="string")
    username = MetricDef(label="User name", type="string")
    password = MetricDef(label="Password", type="string")
    dbname = MetricDef(label="Database name", type="string")
    frequency = MetricDef(label="Frequency", type="string")
    retention = MetricDef(label="Retention", type="string")
    powa_coalesce = MetricDef(label="Powa coalesce", type="string")
    allow_ui_connection = MetricDef(label="Allow UI connection", type="bool")
    snapts = MetricDef(label="Last snapshot", type="string")
    no_err = MetricDef(label="Error", type="bool")
    collector_status = MetricDef(label="Collector Status", type="string")

    query = """SELECT id,
     CASE WHEN s.id = 0 THEN
        '<local>'
     ELSE
        COALESCE(alias,
          s.username || '@' || s.hostname || ':' || s.port || '/' || s.dbname)
     END AS server_alias,
     s.hostname, s.port, s.username,
     CASE WHEN s.password IS NULL THEN '<NULL>' ELSE '********' END AS password,
     s.dbname, s.frequency, s.retention::text AS retention,
     s.powa_coalesce::text AS powa_coalesce, s.allow_ui_connection,
     CASE WHEN coalesce(m.snapts, '-infinity') = '-infinity'::timestamptz THEN
        NULL
     ELSE
        clock_timestamp() - m.snapts
     END::text AS last_snapshot,
     CASE WHEN coalesce(m.snapts, '-infinity') = '-infinity'::timestamptz THEN
        NULL
     ELSE
        m.snapts
     END AS snapts,
     errors IS NULL AS no_err,
     CASE WHEN s.id = 0 THEN
       coalesce(a.val, 'stopped')
     ELSE
       'unknown'
     END AS collector_status
     FROM {powa}.powa_servers s
     LEFT JOIN {powa}.powa_snapshot_metas m ON s.id = m.srvid
     LEFT JOIN (SELECT
        CASE WHEN current_setting('powa.frequency') = '-1' THEN 'disabled'
            ELSE 'running'
        END AS val, application_name
       FROM pg_stat_activity
       WHERE backend_type = 'powa'
     ) a ON a.application_name LIKE 'PoWA - %%'
     ORDER BY 2"""

    def process(self, val, **kwargs):
        val["url"] = self.reverse_url("RemoteConfigOverview", val["id"])
        return val

    def post_process(self, data, **kwargs):
        if len(data["data"]):
            raw = self.notify_collector("WORKERS_STATUS", timeout=1)
            if not raw:
                return data

            line = None
            # get the first correct response only, if multiple answers were
            # returned
            while line is None and len(raw) > 0:
                tmp = raw.pop(0)
                if "OK" in tmp:
                    line = tmp["OK"]

            # nothing correct, give up
            if line is None or line == {}:
                return data

            stats = json.loads(line)

            for row in data["data"]:
                srvid = str(row["id"])
                if srvid in stats:
                    row["collector_status"] = stats[srvid]

        return data


class PgSettingsMetricGroup(MetricGroupDef):
    """
    Metric group for the pg_settings grid.
    """

    name = "pg_settings"
    xaxis = "setting_name"
    data_url = r"/config/(\d+)/pg_settings/"
    axis_type = "category"
    setting_value = MetricDef(label="Value", type="string")
    setting_unit = MetricDef(label="Unit", type="string")
    category_value = MetricDef(label="Category", type="string")
    __query = """
             SELECT name as setting_name, setting as setting_value,
             COALESCE(unit,'') AS setting_unit, category as category_value
             FROM pg_settings
             --WHERE name like 'powa%%'
             ORDER BY name"""
    params = ["server"]

    @property
    def query(self):
        if self.path_args[0] == "0":
            return self.__query
        else:
            # we'll get the data on the foreign server in post_process
            return None

    def post_process(self, data, server, **kwargs):
        # For local server we can return data already retrieved
        if server == "0":
            return data

        values = None

        # Check first if the info is available locally
        if self.has_extension_version(server, "pg_track_settings", "2.0.0"):
            try:
                values = self.execute(
                    """
                        SELECT t.name AS setting_name,
                            t.setting AS setting_value,
                            s.unit AS setting_unit,
                            s.category AS category_value
                        FROM {pg_track_settings}.pg_track_settings(now(),
                                                                   %(srvid)s) t
                        LEFT JOIN pg_catalog.pg_settings s
                            ON s.name = t.name
                        """,
                    params={"srvid": server},
                )

                # If no rows were retrieved, it probably means that
                # pg_tracksettings isn't sampled even if the extension exists.
                # Reset values so we can try to fetch info from the remote
                # server.
                if values.rowcount == 0:
                    values = None
            except Exception:
                # ignore any error, we'll just fallback on remote check
                pass

        if values is None:
            try:
                values = self.execute(self.__query, srvid=server)
            except Exception:
                # ignore any connection or remote execution error
                pass

        if values is not None:
            data = {"data": [self.process(val) for val in values]}
        else:
            data = {
                "data": [],
                "messages": {
                    "alert": [
                        "Could not retrieve PostgreSQL"
                        + " settings "
                        + "on remote server"
                    ]
                },
            }

        return data


class PgStatExtensionsMetricGroup(MetricGroupDef):
    """
    Metric group for the stat extensions grid.
    """

    name = "pg_stat_extensions"
    xaxis = "extname"
    data_url = r"/config/(\d+)/pg_stat_extensions/"
    axis_type = "category"
    available = MetricDef(label="Available", type="bool")
    installed = MetricDef(label="Installed", type="bool")
    handled = MetricDef(label="Sampled", type="bool")
    extversion = MetricDef(label="Version", type="string")
    params = ["server"]

    @property
    def query(self):
        if self.path_args[0] == "0":
            return """SELECT pe.extname, pae.name IS NOT NULL AS available,
                        pae.installed_version IS NOT NULL AS installed,
                    pec.enabled AS handled,
                    coalesce(pae.installed_version , '-') AS extversion
                FROM {powa}.powa_extensions pe
                JOIN {powa}.powa_extension_functions pef USING (extname)
                LEFT JOIN pg_catalog.pg_available_extensions pae
                    ON pae.name = pe.extname
                LEFT JOIN {powa}.powa_extension_config pec
                    ON pec.extname = pe.extname AND pec.srvid = %(server)s
                WHERE operation = 'snapshot'
                """
        else:
            return """SELECT DISTINCT pe.extname,
                        CASE
                            WHEN pec.version IS NOT NULL THEN true
                            ELSE NULL::bool
                        END AS available,
                        NULL::bool AS installed,
                    pec.enabled AS handled,
                    coalesce(pec.version, '-') AS extversion
                FROM {powa}.powa_extensions pe
                JOIN {powa}.powa_extension_functions pef USING (extname)
                LEFT JOIN pg_catalog.pg_available_extensions pae
                    ON pae.name = pe.extname
                LEFT JOIN {powa}.powa_extension_config pec
                    ON pec.extname = pe.extname AND pec.srvid = %(server)s
                WHERE operation = 'snapshot'
                """

    def post_process(self, data, server, **kwargs):
        """
        Get the missing metadata of the extensions on the remote servers if
        needed
        """
        # We already have all the data for the local server
        if server == "0":
            return data

        res = None
        errmsg = None

        extnames = []

        for row in data["data"]:
            extnames.append(row["extname"])

        try:
            res = self.execute(
                """
            SELECT name AS extname, installed_version
            FROM pg_available_extensions
            WHERE name = ANY(%(extnames)s)""",
                srvid=server,
                params={"extnames": extnames},
            )
        except Exception as e:
            # ignore any connection or remote execution error, but keep the
            # error message
            errmsg = str(e)
            pass

        # if we couldn't get any data, send what we have
        if res is None or len(res) == 0:
            data["messages"] = {
                "alert": [
                    "Could not retrieve extensions"
                    + " on remote server: %s" % errmsg
                ]
            }
            return data

        remote_exts = res

        alerts = []
        for ext in data["data"]:
            found = False

            for r in remote_exts:
                if r["extname"] == ext["extname"]:
                    found = True
                    break

            if not found:
                ext["available"] = False
                ext["installed"] = None
            else:
                ext["available"] = True
                ext["installed"] = r["installed_version"] is not None

            if ext["handled"] and ext["installed"] is None:
                alerts.append(ext["extname"])

        if len(alerts) > 0:
            data["messages"] = {
                "alert": [
                    (
                        "%d extensions need to be installed:%s"
                        % (len(alerts), " ".join(alerts))
                    )
                ]
            }

        return data


class PgSupportExtensionsMetricGroup(MetricGroupDef):
    """
    Metric group for the support extensions grid.
    """

    name = "pg_support_extensions"
    xaxis = "extname"
    data_url = r"/config/(\d+)/pg_support_extensions/"
    axis_type = "category"
    available = MetricDef(label="Available", type="bool")
    installed = MetricDef(label="Installed", type="bool")
    extversion = MetricDef(label="Version", type="string")
    params = ["server"]

    @property
    def query(self):
        if self.path_args[0] == "0":
            return """SELECT pe.extname, pae.name IS NOT NULL AS available,
                        pae.installed_version IS NOT NULL AS installed,
                    coalesce(pae.installed_version , '-') AS extversion
                FROM {powa}.powa_extensions pe
                LEFT JOIN {powa}.powa_extension_functions pef USING (extname)
                LEFT JOIN pg_catalog.pg_available_extensions pae
                    ON pae.name = pe.extname
                WHERE operation IS NULL
                """
        else:
            return """SELECT pe.extname,
                        NULL::bool AS available, NULL::bool AS installed,
                        NULL::text AS extversion
                FROM {powa}.powa_extensions pe
                LEFT JOIN {powa}.powa_extension_functions pef USING (extname)
                WHERE operation IS NULL
                """

    def post_process(self, data, server, **kwargs):
        """
        Get the missing metadata of the extensions on the remote servers
        """
        # We already have all the data for the local server
        if server == "0":
            return data

        res = None
        errmsg = None

        extnames = []

        for row in data["data"]:
            extnames.append(row["extname"])

        try:
            res = self.execute(
                """
            SELECT name AS extname, installed_version
            FROM pg_available_extensions
            WHERE name = ANY(%(extnames)s)""",
                srvid=server,
                params={"extnames": extnames},
            )
        except Exception as e:
            # ignore any connection or remote execution error, but keep the
            # error message
            errmsg = str(e)
            pass

        # if we couldn't get any data, send what we have
        if res is None or len(res) == 0:
            data["messages"] = {
                "alert": [
                    "Could not retrieve extensions"
                    + " on remote server: %s" % errmsg
                ]
            }
            return data

        remote_exts = res

        for ext in data["data"]:
            found = False

            for r in remote_exts:
                if r["extname"] == ext["extname"]:
                    found = True
                    break

            if not found:
                ext["available"] = False
                ext["installed"] = None
                ext["extversion"] = "-"
            else:
                ext["available"] = True
                ext["installed"] = r["installed_version"] is not None
                ext["extversion"] = r["installed_version"]

        return data


class PgDbModulesMetricGroup(MetricGroupDef):
    """
    Metric group for the per-db modules grid.
    """

    name = "pg_db_modules"
    xaxis = "db_module"
    data_url = r"/config/(\d+)/pg_db_modules/"
    axis_type = "category"
    dbnames = MetricDef(label="Database restriction", type="string")
    enabled = MetricDef(label="Enabled", type="bool")
    params = ["server"]

    @property
    def query(self):
        return """SELECT db_module, dbnames::text AS dbnames, enabled
            FROM {powa}.powa_db_module_config
            WHERE srvid = %(server)s
            """


class PgCatalogsMetricGroup(MetricGroupDef):
    """
    Metric group for the per-db catalogs grid.
    """

    name = "pg_catalogs"
    xaxis = "datname"
    data_url = r"/config/(\d+)/pg_catalogs/"
    axis_type = "category"
    last_refresh = MetricDef(label="Last refresh", type="string")
    params = ["server"]

    @property
    def query(self):
        # We combine the database list from powa_databases and
        # powa_cat_database in case some databases are missing in one or the
        # other.
        return """SELECT coalesce(pd.datname, pcd.datname) AS datname,
                last_refresh
            FROM {powa}.powa_databases pd
            FULL OUTER JOIN {powa}.powa_catalog_databases pcd
                USING (srvid, oid)
            WHERE srvid = %(server)s
            """


class RepositoryConfigOverview(DashboardPage):
    """
    Dashboard page for configuration page.
    """

    base_url = r"/config/"
    datasources = [PowaServersMetricGroup, AllCollectorsDetail, ServersErrors]
    title = "Configuration"

    def dashboard(self):
        # This COULD be initialized in the constructor, but tornado < 3 doesn't
        # call it
        if getattr(self, "_dashboard", None) is not None:
            return self._dashboard

        self._dashboard = Dashboard(
            "Server list",
            [
                [AllCollectorsDetail],
                [
                    Grid(
                        "Servers",
                        columns=[
                            {
                                "name": "server_alias",
                                "label": "Server",
                                "url_attr": "url",
                                "direction": "ascending",
                            }
                        ],
                        metrics=PowaServersMetricGroup.all(),
                    )
                ],
                [ServersErrors],
            ],
        )
        return self._dashboard


class RemoteConfigOverview(DashboardPage):
    """
    Dashboard page for configuration page.
    """

    base_url = r"/config/(\d+)"
    datasources = [
        PgSettingsMetricGroup,
        PgStatExtensionsMetricGroup,
        PgSupportExtensionsMetricGroup,
        PgDbModulesMetricGroup,
        PgCatalogsMetricGroup,
        CollectorServerDetail,
    ]
    params = ["server"]
    parent = RepositoryConfigOverview
    # title = 'Remote server configuration'

    @classmethod
    def breadcrum_title(cls, handler, param):
        return handler.deparse_srvid(param["server"])

    def dashboard(self):
        # This COULD be initialized in the constructor, but tornado < 3 doesn't
        # call it
        if getattr(self, "_dashboard", None) is not None:
            return self._dashboard

        grids = [
            [
                Grid(
                    "Stats Extensions",
                    columns=[
                        {
                            "name": "extname",
                            "label": "Extension",
                        }
                    ],
                    metrics=PgStatExtensionsMetricGroup.all(),
                ),
                Grid(
                    "Support Extensions",
                    columns=[
                        {
                            "name": "extname",
                            "label": "Extension",
                        }
                    ],
                    metrics=PgSupportExtensionsMetricGroup.all(),
                ),
            ]
        ]

        if self.path_args[0] != "0":
            grids.append(
                [
                    Grid(
                        "Database modules",
                        columns=[
                            {
                                "name": "db_module",
                                "label": "DB module",
                            }
                        ],
                        metrics=PgDbModulesMetricGroup.all(),
                    ),
                    Grid(
                        "Catalogs",
                        columns=[
                            {
                                "name": "datname",
                                "label": "Database",
                            }
                        ],
                        metrics=PgCatalogsMetricGroup.all(),
                    ),
                ]
            )

        grids.append(
            [
                Grid(
                    "PostgreSQL settings",
                    columns=[
                        {
                            "name": "setting_name",
                            "label": "Setting",
                        }
                    ],
                    metrics=PgSettingsMetricGroup.all(),
                )
            ]
        )

        self._dashboard = Dashboard("Configuration overview", grids)
        return self._dashboard



"""
Dashboards for the SLRU page.
"""

from powa.config import ConfigChangesGlobal
from powa.dashboards import (
    Dashboard,
    DashboardPage,
    Graph,
    Grid,
    MetricDef,
    MetricGroupDef,
)
from powa.server import ByAllSlruMetricGroup, ServerOverview
from powa.sql.utils import sum_per_sec
from powa.sql.views_graph import powa_get_slru_sample


class NameSlruMetricGroup(MetricGroupDef):
    """
    Metric group used by pg_stat_slru graph.
    """

    name = "slru_name"
    xaxis = "name"
    data_url = r"/server/(\d+)/metrics/slru/([a-zA-Z]+)"
    blks_zeroed = MetricDef(
        label="Zeroed",
        type="sizerate",
        desc="Number of blocks zeroed during initializations",
    )
    blks_hit = MetricDef(
        label="Hit",
        type="sizerate",
        desc="Number of times disk blocks were found already"
        " in the SLRU, so that a read was not necessary"
        " (this only includes hits in the SLRU, not the"
        " operating system's file system cache)",
    )
    blks_read = MetricDef(
        label="Read",
        type="sizerate",
        desc="Number of disk blocks read for this SLRU",
    )
    blks_written = MetricDef(
        label="Written",
        type="sizerate",
        desc="Number of disk blocks written for this SLRU",
    )
    blks_exists = MetricDef(
        label="Exists",
        type="sizerate",
        desc="Number of blocks checked for existence for this SLRU",
    )
    flushes = MetricDef(
        label="Flushes",
        type="number",
        desc="Number of flushes of dirty data for this SLRU",
    )
    truncates = MetricDef(
        label="Truncates",
        type="number",
        desc="Number of truncates for this SLRU",
    )

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()

        pg_version_num = handler.get_pg_version_num(handler.path_args[0])
        # if we can't connect to the remote server, assume pg13 or more
        if pg_version_num is not None and pg_version_num < 130000:
            return {}
        return base

    @property
    def query(self):
        query = powa_get_slru_sample("name = %(slru)s")

        from_clause = query

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            sum_per_sec("blks_zeroed"),
            sum_per_sec("blks_hit"),
            sum_per_sec("blks_read"),
            sum_per_sec("blks_written"),
            sum_per_sec("blks_exists"),
            sum_per_sec("flushes"),
            sum_per_sec("truncates"),
        ]

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        WHERE sub.mesure_interval != '0 s'
        GROUP BY sub.srvid, sub.ts, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols),
            from_clause=from_clause,
        )


class ByAllSlruMetricGroup2(ByAllSlruMetricGroup):
    """
    Metric group used by pg_stat_slru grid.

    There isn't a lot of SLRU, so we want to display all lines even in the
    per-SLRU view so users can easily navigate to another one.  Neither the
    backend nor the frontend can handle a metric group used in different
    dashboard, so duplicate it with the bare minimum different information to
    make it work.  It's not a big problem since this is a different page, so we
    won't end up running the same query twice.
    """

    name = "slru_by_name2"
    data_url = r"/server/(\d+)/metrics/slru_by_name2/([a-zA-Z]+)"


class ByNameSlruOverview(DashboardPage):
    """
    Per SLRU Dashboard page.
    """

    base_url = r"/server/(\d+)/metrics/slru/([a-zA-Z]+)/overview/"
    datasources = [NameSlruMetricGroup, ByAllSlruMetricGroup2]
    params = ["server", "slru"]
    parent = ServerOverview
    title = 'Activity for "%(slru)s" SLRU'
    timeline = ConfigChangesGlobal
    timeline_params = ["server"]

    def dashboard(self):
        # This COULD be initialized in the constructor, but tornado < 3 doesn't
        # call it
        if getattr(self, "_dashboard", None) is not None:
            return self._dashboard

        graphs = [
            [
                Graph(
                    "%(slru)s SLRU (per second)",
                    metrics=NameSlruMetricGroup.all(self),
                )
            ],
            [
                Grid(
                    "All SLRUs",
                    columns=[
                        {
                            "name": "name",
                            "label": "SLRU name",
                            "url_attr": "url",
                        }
                    ],
                    metrics=ByAllSlruMetricGroup2.all(self),
                )
            ],
        ]

        self._dashboard = Dashboard(self.title, graphs)
        return self._dashboard



"""
Index page presenting the list of available servers.
"""

from powa.dashboards import (
    Dashboard,
    DashboardPage,
    Grid,
    MetricDef,
    MetricGroupDef,
)


class OverviewMetricGroup(MetricGroupDef):
    """
    Metric group used by the "all servers" grid
    """

    name = "all_servers"
    xaxis = "srvid"
    axis_type = "category"
    data_url = r"/server/all_servers/"
    hostname = MetricDef(label="Hostname", type="text")
    port = MetricDef(label="Port", type="text")
    version = MetricDef(label="Version", type="text")

    @property
    def query(self):
        sql = """SELECT id AS srvid,
                CASE WHEN id = 0 THEN
                   '<local>'
                ELSE
                   COALESCE(alias, hostname || ':' || port)
                END AS alias,
                CASE WHEN id = 0 THEN %(host)s ELSE hostname END as hostname,
                CASE WHEN id = 0 THEN %(port)s ELSE port END AS port,
                CASE WHEN id = 0 THEN set.setting
                    ELSE s.version::text
                END AS version
                FROM {powa}.powa_servers s
                LEFT JOIN pg_settings set ON set.name = 'server_version'
                    AND s.id = 0"""

        return (sql, {"host": self.current_host, "port": self.current_port})

    def process(self, val, **kwargs):
        val["url"] = self.reverse_url("ServerOverview", val["srvid"])
        return val


class Overview(DashboardPage):
    """
    Overview dashboard page.
    """

    base_url = r"/server/"
    datasources = [OverviewMetricGroup]
    title = "All servers"

    def dashboard(self):
        # This COULD be initialized in the constructor, but tornado < 3 doesn't
        # call it
        if getattr(self, "_dashboard", None) is not None:
            return self._dashboard

        dashes = [
            [
                Grid(
                    "All servers",
                    columns=[
                        {
                            "name": "alias",
                            "label": "Instance",
                            "url_attr": "url",
                            "direction": "descending",
                        }
                    ],
                    metrics=OverviewMetricGroup.all(),
                )
            ]
        ]

        self._dashboard = Dashboard("All servers", dashes)
        return self._dashboard

    @classmethod
    def get_childmenu(cls, handler, params):
        from powa.server import ServerOverview

        children = []
        for s in list(handler.servers):
            new_params = params.copy()
            new_params["server"] = s[0]
            entry = ServerOverview.get_selfmenu(handler, new_params)
            entry.title = s[2]
            children.append(entry)
        return children



from __future__ import print_function

"""
Powa main application.
"""

import os
import re

__VERSION__ = "5.1.1"

ver_tmp = re.sub("(alpha|beta|dev)[0-9]*", "", __VERSION__)
__VERSION_NUM__ = [int(part) for part in (ver_tmp.split("."))]

POWA_ROOT = os.path.dirname(__file__)

# Import from powa.options must go before tornado.options
from powa.options import parse_options  # noqa: I001
from tornado.options import options
from tornado.web import Application
from tornado.web import URLSpec as U

from powa import ui_methods, ui_modules
from powa.collector import (
    CollectorDbCatRefreshHandler,
    CollectorForceSnapshotHandler,
    CollectorReloadHandler,
)
from powa.config import (
    RemoteConfigOverview,
    RepositoryConfigOverview,
)
from powa.database import DatabaseOverview, DatabaseSelector
from powa.framework import AuthHandler
from powa.function import FunctionOverview
from powa.io import (
    ByBackendTypeIoOverview,
    ByContextIoOverview,
    ByObjIoOverview,
)
from powa.overview import Overview
from powa.qual import QualOverview
from powa.query import QueryOverview
from powa.server import ServerOverview, ServerSelector
from powa.slru import ByNameSlruOverview
from powa.user import LoginHandler, LogoutHandler
from powa.wizard import IndexSuggestionHandler


class IndexHandler(AuthHandler):
    """
    Handler for the main page.
    """

    def get(self):
        return self.redirect(options.index_url)


def make_app(**kwargs):
    """
    Parse the config file and instantiate a tornado app.
    """
    parse_options()

    URLS = [
        U(r"%slogin/" % options.url_prefix, LoginHandler, name="login"),
        U(r"%slogout/" % options.url_prefix, LogoutHandler, name="logout"),
        U(
            r"%sreload_collector/" % options.url_prefix,
            CollectorReloadHandler,
            name="reload_collector",
        ),
        U(
            r"%sforce_snapshot/(\d+)" % options.url_prefix,
            CollectorForceSnapshotHandler,
            name="force_snapshot",
        ),
        U(
            r"%srefresh_db_cat/" % options.url_prefix,
            CollectorDbCatRefreshHandler,
            name="refresh_db_cat",
        ),
        U(
            r"%sserver/select" % options.url_prefix,
            ServerSelector,
            name="server_selector",
        ),
        U(
            r"%sdatabase/select" % options.url_prefix,
            DatabaseSelector,
            name="database_selector",
        ),
        U(r"%s" % options.url_prefix, IndexHandler, name="index"),
        U(
            r"%sserver/(\d+)/database/([^\/]+)/suggest/" % options.url_prefix,
            IndexSuggestionHandler,
            name="index_suggestion",
        ),
    ]

    for dashboard in (
        Overview,
        ServerOverview,
        DatabaseOverview,
        QueryOverview,
        QualOverview,
        FunctionOverview,
        RepositoryConfigOverview,
        RemoteConfigOverview,
        ByBackendTypeIoOverview,
        ByObjIoOverview,
        ByContextIoOverview,
        ByNameSlruOverview,
    ):
        URLS.extend(dashboard.url_specs(options.url_prefix))

    _cls = Application
    if "legacy_wsgi" in kwargs:
        from tornado.wsgi import WSGIApplication

        _cls = WSGIApplication

    return _cls(
        URLS,
        ui_modules=ui_modules,
        ui_methods=ui_methods,
        login_url=("%slogin/" % options.url_prefix),
        static_path=os.path.join(POWA_ROOT, "static"),
        static_url_prefix=("%sstatic/" % options.url_prefix),
        cookie_secret=options.cookie_secret,
        template_path=os.path.join(POWA_ROOT, "templates"),
        **kwargs,
    )



"""
Py2/3 compatibility layer

Mostly copied straight from six:
    http://pypi.python.org/pypi/six/

"""

from __future__ import absolute_import

import json
import psycopg2
from psycopg2 import extensions

# If psycopg2 < 2.5, register json type
psycopg2_version = tuple(psycopg2.__version__.split(" ")[0].split("."))
if psycopg2_version < ("2", "5"):
    JSON_OID = 114
    newtype = extensions.new_type(
        (JSON_OID,), "JSON", lambda data, cursor: json.loads(data)
    )
    extensions.register_type(newtype)


def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""

    # This requires a bit of explanation: the basic idea is to make a dummy
    # metaclass for one level of class instantiation that replaces itself with
    # the actual metaclass.
    class metaclass(meta):
        """The actual metaclass."""

        def __new__(cls, name, _, d):
            return meta(name, bases, d)

    return type.__new__(metaclass, "temporary_class", (), {})


class classproperty(object):
    """
    A descriptor similar to property, but using the class.
    """

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)



"""
Utilities for commonly used SQL constructs.
"""

import re
import sys
from collections import defaultdict, namedtuple
from powa.json import JSONizable

TOTAL_MEASURE_INTERVAL = """
extract( epoch from
    CASE WHEN min(total_mesure_interval) = '0 second'
        THEN '1 second'::interval
    ELSE min(total_mesure_interval) END)
"""


def unprepare(sql):
    if sql.startswith("PREPARE"):
        sql = re.sub("PREPARE.*AS", "", sql)
    sql = re.sub("\$\d+", "?", sql)
    return sql


def format_jumbled_query(sql, params):
    sql = unprepare(sql)
    it = iter(params)
    try:
        sql = re.sub("\?", lambda val: next(it), sql)
    except StopIteration:
        pass
    return sql


RESOLVE_OPNAME = """
SELECT json_object_agg(oid, value)
    FROM (

    SELECT oproid as oid, json_build_object(
        'name', oprname,
        'amop',
        coalesce(
            json_object_agg(am, opclass_oids::jsonb
            ORDER BY am)
                      FILTER (WHERE am is NOT NULL)),
        'amop_names',
         coalesce(
            json_object_agg(
                      amname,
                      opclass_names ORDER BY am)
                      FILTER (WHERE am is NOT NULL))) as value
    FROM
    (
        SELECT oprname, pg_operator.oid as oproid,
            pg_am.oid as am, to_json(array_agg(distinct c.oid)) as opclass_oids,
            amname,
        to_json(array_agg(distinct CASE
                      WHEN opcdefault IS TRUE THEN ''
                      WHEN opcdefault IS FALSE THEN opcname
                      ELSE NULL END)) as opclass_names
        FROM
        pg_operator
        LEFT JOIN pg_amop amop ON amop.amopopr = pg_operator.oid
        LEFT JOIN pg_am ON amop.amopmethod = pg_am.oid AND pg_am.amname != 'hash'
        LEFT JOIN pg_opfamily f ON f.opfmethod = pg_am.oid AND amop.amopfamily = f.oid
        LEFT JOIN pg_opclass c ON c.opcfamily = f.oid
        WHERE pg_operator.oid in %(oid_list)s
        GROUP BY pg_operator.oid, oprname, pg_am.oid, amname
    ) by_am
    GROUP BY oproid, oprname
    ) detail
"""

RESOLVE_ATTNAME = """
    SELECT json_object_agg(attrelid || '.'|| attnum, value)
    FROM (
    SELECT attrelid, attnum, json_build_object(
        'relname', quote_ident(relname),
        'attname', quote_ident(attname),
        'nspname', quote_ident(nspname),
        'n_distinct', COALESCE(stadistinct, 0),
        'null_frac', stanullfrac,
        'most_common_values', CASE
            WHEN s.stakind1 = 1 THEN s.stavalues1
            WHEN s.stakind2 = 1 THEN s.stavalues2
            WHEN s.stakind3 = 1 THEN s.stavalues3
            WHEN s.stakind4 = 1 THEN s.stavalues4
            WHEN s.stakind5 = 1 THEN s.stavalues5
            ELSE NULL::anyarray
        END,
        'table_liverows', pg_stat_get_live_tuples(c.oid)
    ) as value
    FROM pg_attribute a
    INNER JOIN pg_class c on c.oid = a.attrelid
    INNER JOIN pg_namespace n ON n.oid = c.relnamespace
    LEFT JOIN pg_statistic s ON s.starelid = c.oid
                       AND s.staattnum = a.attnum
    WHERE (attrelid, attnum) IN %(att_list)s
    ) detail
"""


class ResolvedQual(JSONizable):
    def __init__(
        self,
        nspname,
        relname,
        attname,
        opname,
        amops,
        n_distinct=0,
        most_common_values=None,
        null_frac=None,
        example_values=None,
        eval_type=None,
        relid=None,
        attnum=None,
    ):
        self.nspname = nspname
        self.relname = relname
        self.attname = attname
        self.opname = opname
        self.amops = amops or {}
        self.n_distinct = n_distinct
        self.most_common_values = most_common_values
        self.null_frac = null_frac
        self.example_values = example_values or []
        self.eval_type = eval_type
        self.relid = relid
        self.attnum = attnum

    def __str__(self):
        return "{}.{} {} ?".format(self.relname, self.attname, self.opname)

    @property
    def distinct_values(self):
        if self.n_distinct == 0:
            return None
        elif self.n_distinct > 0:
            return "{}".format(self.n_distinct)
        else:
            return "{} %".format(abs(self.n_distinct) * 100)

    def to_json(self):
        base = super(ResolvedQual, self).to_json()
        base["label"] = str(self)
        base["distinct_values"] = self.distinct_values
        return base


class ComposedQual(JSONizable):
    def __init__(
        self,
        nspname=None,
        relname=None,
        avg_filter=None,
        filter_ratio=None,
        occurences=None,
        execution_count=None,
        table_liverows=None,
        qualid=None,
        relid=None,
        queries=None,
        queryids=None,
    ):
        super(ComposedQual, self).__init__()
        self.qualid = qualid
        self.relname = relname
        self.nspname = nspname
        self.avg_filter = avg_filter
        self.filter_ratio = filter_ratio
        self.execution_count = execution_count
        self.occurences = occurences
        self.table_liverows = table_liverows
        self.relid = relid
        self.queries = queries or []
        self.queryids = queryids or []
        self._quals = []

    def append(self, element):
        if not isinstance(element, ResolvedQual):
            raise ValueError(
                ("ComposedQual elements must be instances of ", "ResolvedQual")
            )
        self._quals.append(element)

    def __iter__(self):
        return self._quals.__iter__()

    def __str__(self):
        return " AND ".join(str(v) for v in self._quals)

    @property
    def where_clause(self):
        return "WHERE {}".format(self)

    def to_json(self):
        base = super(ComposedQual, self).to_json()
        base["quals"] = self._quals
        base["where_clause"] = self.where_clause
        return base


def resolve_quals(conn, quallist, attribute="quals"):
    """
    Resolve quals definition (as dictionary coming from a to_json(quals)
    sql query.

    Arguments:
        conn: a connection to the database against which the qual was executed
        quallist: an iterable of rows, each storing quals in the attributes
        attribute: the attribute containing the qual list itself in each row
    Returns:
        a list of ComposedQual objects
    """
    operator_to_look = set()
    attname_to_look = set()
    operators = {}
    attnames = {}
    for row in quallist:
        values = row[attribute]
        if not isinstance(values, list):
            values = [values]
        for v in values:
            operator_to_look.add(v["opno"])
            attname_to_look.add((v["relid"], v["attnum"]))
    if operator_to_look:
        cur = conn.cursor()
        cur.execute(RESOLVE_OPNAME, {"oid_list": tuple(operator_to_look)})
        operators = cur.fetchone()[0]
        cur.close()
    if attname_to_look:
        cur = conn.cursor()
        cur.execute(RESOLVE_ATTNAME, {"att_list": tuple(attname_to_look)})
        attnames = cur.fetchone()[0]
    new_qual_list = []
    for row in quallist:
        row = dict(row)
        newqual = ComposedQual(
            occurences=row["occurences"],
            execution_count=row["execution_count"],
            avg_filter=row["avg_filter"],
            filter_ratio=row["filter_ratio"],
            qualid=row["qualid"],
            queries=row.get("queries"),
            queryids=row.get("queryids"),
        )
        new_qual_list.append(newqual)
        values = [v for v in row[attribute] if v["relid"] != "0"]
        if not isinstance(values, list):
            values = [values]
        for v in values:
            attname = attnames["{}.{}".format(v["relid"], v["attnum"])]
            if newqual.relname is not None:
                if newqual.relname != attname["relname"]:
                    raise ValueError(
                        "All individual qual parts should be on the "
                        "same relation"
                    )
            else:
                newqual.relname = attname["relname"]
                newqual.nspname = attname["nspname"]
                newqual.relid = v["relid"]
                newqual.table_liverows = attname["table_liverows"]
            newqual.append(
                ResolvedQual(
                    nspname=attname["nspname"],
                    relname=attname["relname"],
                    attname=attname["attname"],
                    opname=operators[v["opno"]]["name"],
                    amops=operators[v["opno"]]["amop_names"],
                    n_distinct=attname["n_distinct"],
                    most_common_values=attname["most_common_values"],
                    null_frac=attname["null_frac"],
                    eval_type=v["eval_type"],
                    relid=v["relid"],
                    attnum=v["attnum"],
                )
            )
    return new_qual_list


Plan = namedtuple(
    "Plan",
    (
        "title",
        "values",
        "query",
        "plan",
        "filter_ratio",
        "exec_count",
        "occurences",
    ),
)


def qual_constants(
    srvid, type, filter_clause, queries=None, quals=None, top=1
):
    """
    filter_clause is a plain string corresponding to the list of predicates to
    apply in the main query.

    queries and quals are optional list of respectively queryid and qualid to
    filter on.
    """
    orders = {
        "most_executed": "8 DESC",
        "least_filtering": "9",
        "most_filtering": "9 DESC",
        "most_used": "6 DESC",
    }
    if type not in (
        "most_executed",
        "most_filtering",
        "least_filtering",
        "most_used",
    ):
        return

    query_subfilter = ""
    query_filter = ""
    qual_subfilter = ""
    qual_filter = ""

    if queries is not None:
        query_subfilter = "AND queryid IN ({})".format(queries)
        query_filter = "AND s.queryid IN ({})".format(queries)

    if quals is not None:
        qual_subfilter = "AND qualid IN ({})".format(quals)
        qual_filter = "AND qnc.qualid IN ({})".format(quals)

    base = """
    (
    WITH sample AS (
    SELECT s.srvid, query, s.queryid, qn.qualid, quals as quals,
                constants,
                sum(occurences) as occurences,
                sum(execution_count) as execution_count,
                sum(nbfiltered) as nbfiltered,
                CASE WHEN sum(execution_count) = 0 THEN 0 ELSE sum(nbfiltered) / sum(execution_count) END AS filter_ratio
        FROM {{powa}}.powa_statements s
        JOIN {{powa}}.powa_databases d ON d.oid = s.dbid AND d.srvid = s.srvid
        JOIN {{powa}}.powa_qualstats_quals qn ON s.queryid = qn.queryid AND s.srvid = qn.srvid
        JOIN (
            SELECT *
            FROM {{powa}}.powa_qualstats_constvalues_history qnc
            WHERE srvid = {srvid}
            {query_subfilter}
            {qual_subfilter}
              AND coalesce_range && tstzrange(%(from)s, %(to)s)
            UNION ALL
            SELECT *
            FROM {{powa}}.powa_qualstats_aggregate_constvalues_current(%(server)s, %(from)s, %(to)s)
            WHERE srvid = {srvid}
            {query_subfilter}
            {qual_subfilter}
        ) qnc ON qnc.srvid = s.srvid AND qn.qualid = qnc.qualid AND qn.queryid = qnc.queryid,
        LATERAL
                unnest({qual_type}) as t(constants,occurences, execution_count, nbfiltered)
        WHERE {filter}
        {query_filter}
        {qual_filter}
        AND s.srvid = {srvid}
        GROUP BY s.srvid, qn.qualid, quals, constants, s.queryid, query
        ORDER BY {order}
        LIMIT {top_value}
    )
    SELECT srvid, query, queryid, qualid, quals, constants as constants,
                occurences as occurences,
                nbfiltered as nbfiltered,
                execution_count as execution_count,
                filter_ratio as filter_ratio,
                row_number() OVER (ORDER BY execution_count desc NULLS LAST) AS rownumber
        FROM sample
    ORDER BY rownumber
    LIMIT {top_value}
    ) {qual_type}
    """.format(
        qual_type=type,
        filter=filter_clause,
        query_subfilter=query_subfilter,
        query_filter=query_filter,
        qual_subfilter=qual_subfilter,
        qual_filter=qual_filter,
        order=orders[type],
        top_value=top,
        srvid=srvid,
    )

    query = "SELECT * FROM " + base

    return query


def quote_ident(name):
    return '"' + name + '"'


def get_plans(cls, server, database, query, all_vals):
    plans = []
    for key in (
        "most filtering",
        "least filtering",
        "most executed",
        "most used",
    ):
        vals = all_vals[key]
        query = format_jumbled_query(query, vals["constants"])
        plan = "N/A"
        try:
            sqlQuery = "EXPLAIN {}".format(query)
            result = cls.execute(
                sqlQuery, srvid=server, database=database, remote_access=True
            )
            plan = "\n".join(v["QUERY PLAN"] for v in result)
        except Exception as e:
            plan = "ERROR: %r" % e
            pass
        plans.append(
            Plan(
                key,
                vals["constants"],
                query,
                plan,
                vals["filter_ratio"],
                vals["execution_count"],
                vals["occurences"],
            )
        )
    return plans


def get_unjumbled_query(
    ctrl, srvid, database, queryid, _from, _to, kind="most executed"
):
    """
    From a queryid, build a query string ready to be executed.

    Gather a jumbled query from powa_statements, then try to denormalized it
    using stored const values, by its kind (least/most filtering, most common).

    This function can return None if the query has not been gathered by powa,
    or a partially or fully normalized query, depending on const values has
    been found and/or the SELECT clause has been normalized
    """

    rs = list(
        ctrl.execute(
            """
        SELECT query
        FROM {powa}.powa_statements
        WHERE srvid= %(srvid)s
        AND queryid = %(queryid)s LIMIT 1
    """,
            params={"srvid": srvid, "queryid": queryid},
        )
    )[0]
    normalized_query = rs["query"]
    values = qualstat_get_figures(
        ctrl, srvid, database, _from, _to, queries=[queryid]
    )

    if values is None:
        unprepared = unprepare(normalized_query)
        if unprepared != normalized_query:
            return None

    # Try to inject values
    sql = format_jumbled_query(
        normalized_query, values[kind].get("constants", [])
    )
    return sql


def get_any_sample_query(ctrl, srvid, database, queryid, _from, _to):
    """
    From a queryid, get a non normalized query.

    If pg_qualstats is available and recent enough, try to retrieve a randomly
    chosen non normalized query, which is fast, trying to avoid EXPLAIN
    queries.

    If this fail, fallback get_unjumbled_query, with "most executed" const
    values.
    """
    has_pgqs = ctrl.has_extension_version(srvid, "pg_qualstats", "0.0.7")
    example_query = None
    if has_pgqs:
        rows = ctrl.execute(
            """
            WITH s(max, v) AS (
                SELECT
                (SELECT setting::int
                    FROM pg_catalog.pg_settings
                    WHERE name = 'track_activity_query_size'
                ), {pg_qualstats}.pg_qualstats_example_query(%(queryid)s)
            )
            SELECT max, v, octet_length(v) AS len
            FROM s
            WHERE v NOT ILIKE '%%EXPLAIN%%'
            LIMIT 1
        """,
            params={"queryid": queryid},
            srvid=srvid,
            remote_access=True,
        )
        if rows is not None and len(rows) > 0:
            # Ignore the query if it looks like it was truncated
            if rows[0]["len"] < (rows[0]["max"] - 1):
                example_query = rows[0]["v"]
        if example_query is not None:
            unprepared = unprepare(example_query)
            if example_query == unprepared:
                return example_query
    return get_unjumbled_query(
        ctrl, srvid, database, queryid, _from, _to, "most executed"
    )


def qualstat_get_figures(
    cls, srvid, database, tsfrom, tsto, queries=None, quals=None
):
    condition = """datname = %(database)s
            AND coalesce_range && tstzrange(%(from)s, %(to)s)"""

    if queries is not None:
        queries_str = ",".join(str(q) for q in queries)

    cols = [
        "most_filtering.quals",
        "most_filtering.query",
        'to_json(most_filtering) as "most filtering"',
        'to_json(least_filtering) as "least filtering"',
        'to_json(most_executed) as "most executed"',
        'to_json(most_used) as "most used"',
    ]

    sql = """SELECT {cols}
    FROM ({most_filtering}) AS most_filtering
    JOIN ({least_filtering}) AS least_filtering USING (rownumber)
    JOIN ({most_executed}) AS most_executed USING (rownumber)
    JOIN ({most_used}) AS most_used USING (rownumber)""".format(
        cols=", ".join(cols),
        most_filtering=qual_constants(
            srvid, "least_filtering", condition, queries_str, quals
        ),
        least_filtering=qual_constants(
            srvid, "least_filtering", condition, queries_str, quals
        ),
        most_executed=qual_constants(
            srvid, "most_executed", condition, queries_str, quals
        ),
        most_used=qual_constants(
            srvid, "most_used", condition, queries_str, quals
        ),
    )

    params = {
        "server": srvid,
        "database": database,
        "from": tsfrom,
        "to": tsto,
        "queryids": queries,
    }
    quals = cls.execute(sql, params=params)

    if len(quals) == 0:
        return None

    row = quals[0]

    return row


class HypoPlan(JSONizable):
    def __init__(
        self, baseplan, basecost, hypoplan, hypocost, query, indexes=None
    ):
        self.baseplan = baseplan
        self.basecost = basecost
        self.hypoplan = hypoplan
        self.hypocost = hypocost
        self.query = query
        self.indexes = indexes or []

    @property
    def gain_percent(self):
        return round(
            100 - float(self.hypocost) * 100 / float(self.basecost), 2
        )

    def to_json(self):
        base = super(HypoPlan, self).to_json()
        base["gain_percent"] = self.gain_percent
        return base


class HypoIndex(JSONizable):
    def __init__(self, nspname, relname, amname, composed_qual=None):
        self.nspname = nspname
        self.relname = relname
        self.qual = composed_qual
        self.amname = amname
        self.name = None
        self._ddl = None
        self.__setattr__ = self.__setattr
        self._update_ddl()

    def _update_ddl(self):
        if "btree" == self.amname:
            attrs = []
            for qual in self.qual:
                if qual.attname not in attrs:
                    attrs.append(qual.attname)
            # Qual resolution is responsible for quoting all identifiers
            super(HypoIndex, self).__setattr__(
                "_ddl",
                """CREATE INDEX ON {nsp}.{rel}({attrs})""".format(
                    nsp=self.nspname, rel=self.relname, attrs=",".join(attrs)
                ),
            )

    def __setattr(self, name, value):
        super(HypoIndex, self).__setattr__(name, value)
        # Only btree is supported right now
        if name in ("amname", "nspname", "relname", "composed_qual"):
            self._update_ddl()

    @property
    def ddl(self):
        return self._ddl

    @property
    def hypo_ddl(self):
        ddl = self.ddl
        if ddl is not None:
            return (
                "SELECT indexname"
                + " FROM {hypopg}.hypopg_create_index(%(sql)s)",
                {"sql": ddl},
            )
        return (None, None)

    def to_json(self):
        base = super(HypoIndex, self).to_json()
        base["ddl"] = self.ddl
        return base


def possible_indexes(composed_qual, order=()):
    by_am = defaultdict(list)

    def sorter(qual):
        attnum = qual.attnum
        if attnum in order:
            return order.index(attnum)
        else:
            # - attnum to guarantee stable sort
            return sys.maxsize - attnum

    for qual in sorted(composed_qual, key=sorter):
        for am in qual.amops.keys():
            by_am[am].append(qual)
    indexes = []
    for am, quals in by_am.items():
        base = quals[0]
        indexes.append(HypoIndex(base.nspname, base.relname, am, quals))
    return indexes


def get_hypoplans(cur, query, indexes=None):
    """
    With a connection to a database where hypothetical indexes
    have already been created, request two plans for each query:
        - one with hypothetical indexes
        - one without hypothetical indexes

    Arguments:
        conn: a connection to the target database
        queries: a list of sql queries, already formatted with values
        indexes: a list of hypothetical index names to look for in the plan
    """
    # Early exit if no query was provided
    if query is None:
        return None

    indexes = indexes or []
    # Escape literal '%'
    query = query.replace("%", "%%")
    cur.execute("SET hypopg.enabled = off")
    try:
        cur.execute("SAVEPOINT hypo")
        cur.execute("EXPLAIN {}".format(query))
        baseplan = "\n".join(v[0] for v in cur.fetchall())
        cur.execute("RELEASE hypo")
    except Exception as e:
        cur.execute("ROLLBACK TO hypo")
        raise e

    cur.execute("SET hypopg.enabled = on")
    try:
        cur.execute("SAVEPOINT hypo")
        cur.execute("EXPLAIN {}".format(query))
        hypoplan = "\n".join(v[0] for v in cur.fetchall())
        cur.execute("RELEASE hypo")
    except Exception as e:
        cur.execute("ROLLBACK TO hypo")
        raise e

    COST_RE = "(?<=\.\.)\d+\.\d+"
    m = re.search(COST_RE, baseplan)
    if m:
        basecost = float(m.group(0))
    else:
        basecost = None
    m = re.search(COST_RE, hypoplan)
    if m:
        hypocost = float(m.group(0))
    else:
        hypocost = None
    used_indexes = []
    for ind in indexes:
        if ind.name is not None and ind.name in hypoplan:
            used_indexes.append(ind)
    return HypoPlan(
        baseplan, basecost, hypoplan, hypocost, query, used_indexes
    )



from __future__ import unicode_literals

from powa import __VERSION_NUM__
from powa.framework import BaseHandler
from tornado.options import options


class LoginHandler(BaseHandler):
    def get(self):
        self._status_code = 403
        return self.render("login.html", title="Login")

    def post(self, *args, **kwargs):
        user = self.get_argument("user")
        password = self.get_argument("password")
        server = self.get_argument("server")
        expires_days = options.cookie_expires_days
        if expires_days == 0:
            expires_days = None

        try:
            self.connect(user=user, password=password, server=server)
        except Exception as e:
            self.flash("Auth failed", "alert")
            self.logger.error("Error: %r", e)
            self.get()
            return
        # Check that the database is correctly installed
        version = self.get_powa_version(
            user=user, password=password, server=server
        )
        if version is None:
            self.flash(
                "PoWA is not installed on your target database. "
                "You should check your installation.",
                "alert",
            )
            self.redirect(self.url_prefix)
        # Major.Minor version should be the same
        if version[0:2] != __VERSION_NUM__[0:2]:
            self.flash(
                "Unable to connect: powa-archivist version %s.X does not match powa-web version %s.X"
                % (
                    ".".join(str(x) for x in version[0:2]),
                    ".".join(str(x) for x in __VERSION_NUM__[0:2]),
                ),
                "alert",
            )
            self.redirect(self.url_prefix)
        self.set_secure_cookie("user", user, expires_days=expires_days)
        self.set_secure_cookie("password", password, expires_days=expires_days)
        self.set_secure_cookie("server", server, expires_days=expires_days)
        self.redirect(self.get_argument("next", self.url_prefix))


class LogoutHandler(BaseHandler):
    def get(self):
        self.clear_all_cookies()
        return self.redirect(self.url_prefix)
